from typing import Any, Dict, Tuple, Union

import optuna
import pandas as pd

from bluecast.config.training_config import (
    TrainingConfig,
    XgboostFinalParamConfig,
    XgboostRegressionFinalParamConfig,
    XgboostTuneParamsConfig,
    XgboostTuneParamsRegressionConfig,
)
from bluecast.general_utils.general_utils import check_gpu_support, log_sampling


def update_params_based_on_tree_method(
    param: Dict[str, Any],
    trial: optuna.Trial,
    xgboost_params: Union[XgboostTuneParamsConfig, XgboostTuneParamsRegressionConfig],
) -> Dict[str, Any]:
    """Update parameters based on tree method."""

    param["tree_method"] = trial.suggest_categorical(
        "tree_method", xgboost_params.tree_method
    )

    param["booster"] = trial.suggest_categorical("booster", xgboost_params.booster)
    if param["booster"] == "gbtree":
        param["grow_policy"] = trial.suggest_categorical(
            "grow_policy", xgboost_params.grow_policy
        )
    elif param["booster"] == "gblinear":
        del param["max_depth"]
        del param["min_child_weight"]
        del param["subsample"]
        del param["colsample_bytree"]
        del param["colsample_bylevel"]
        del param["gamma"]
        del param["tree_method"]
    return param


def get_params_based_on_device(
    conf_training: TrainingConfig,
    conf_params_xgboost: Union[
        XgboostFinalParamConfig, XgboostRegressionFinalParamConfig
    ],
    conf_xgboost: Union[XgboostTuneParamsConfig, XgboostTuneParamsRegressionConfig],
) -> Dict[str, Any]:
    """Get parameters based on available or chosen device."""
    if conf_training.autotune_on_device in ["auto", "gpu"]:
        train_on = check_gpu_support()
        conf_params_xgboost.params["device"] = train_on["device"]
        if "exact" in conf_xgboost.tree_method and conf_params_xgboost.params[
            "device"
        ] in ["gpu", "cuda"]:
            conf_xgboost.tree_method.remove("exact")
    else:
        train_on = {"tree_method": "exact", "device": "cpu"}
    return train_on


def update_params_with_best_params(
    param: Dict[str, Any],
    best_params: Union[XgboostTuneParamsConfig, XgboostTuneParamsRegressionConfig],
) -> Dict[str, Any]:
    """Update parameters based on best parameters after tuning."""

    params_to_check = ["tree_method", "booster", "grow_policy"]
    for param_name in params_to_check:
        if param_name in best_params:
            param[param_name] = best_params[param_name]
    return param


def update_hyperparam_space_after_nth_trial(
    trial: optuna.Trial,
    conf_xgboost: Union[XgboostTuneParamsConfig, XgboostTuneParamsRegressionConfig],
    nth_trial: int = 25,
) -> Union[XgboostTuneParamsConfig, XgboostTuneParamsRegressionConfig]:
    eta_min_before = conf_xgboost.eta_min

    if trial.number % nth_trial * 2 == 0 and eta_min_before == 5e-2:
        conf_xgboost.eta_min = 1e-3
        conf_xgboost.eta_max = 0.3
        conf_xgboost.sub_sample_min = 0.5
        conf_xgboost.min_child_weight_max = 100.0

    if trial.number % nth_trial == 0 and eta_min_before == 1e-3:
        conf_xgboost.eta_min = 5e-2
        conf_xgboost.eta_max = 0.25
        conf_xgboost.sub_sample_min = 1.0
        conf_xgboost.min_child_weight_max = 10.0
        conf_xgboost.col_sample_by_level_min = 0.5

    return conf_xgboost


def sample_data(
    x_train: pd.DataFrame,
    x_test: pd.DataFrame,
    y_train: pd.Series,
    y_test: pd.Series,
    conf_training: TrainingConfig,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.Series, pd.Series]:
    if conf_training.sample_data_during_tuning:
        nb_samples_train = log_sampling(
            len(x_train.index),
            alpha=conf_training.sample_data_during_tuning_alpha,
        )
        nb_samples_test = log_sampling(
            len(x_test.index),
            alpha=conf_training.sample_data_during_tuning_alpha,
        )

        x_train = x_train.sample(
            nb_samples_train, random_state=conf_training.global_random_state
        )
        y_train = y_train.loc[x_train.index]
        x_test = x_test.sample(
            nb_samples_test, random_state=conf_training.global_random_state
        )
        y_test = y_test.loc[x_test.index]
    return x_train, x_test, y_train, y_test



from typing import Dict, Union

import numpy as np
import pandas as pd


def convert_expected_effectiveness_nonconformity_input_types(
    y_hat: Union[np.ndarray, pd.Series, pd.DataFrame]
) -> np.ndarray:
    if isinstance(y_hat, pd.Series):
        y_hat = y_hat.values
        return y_hat
    elif isinstance(y_hat, np.ndarray):
        return y_hat
    elif isinstance(y_hat, pd.DataFrame):
        y_hat = y_hat.values
        return y_hat
    else:
        raise ValueError(
            f"Please provide a numpy array or Pandas Series rather than a {type(y_hat)}."
        )


def one_c(y_hat: Union[np.ndarray, pd.DataFrame, pd.Series]):
    """
    Calculate proportion of singleton sets among all prediction sets.
    :param y_hat: Predicted probabilities of shape (n_samples, 1) where each row is a set of classes.
    """
    if isinstance(y_hat, np.ndarray):
        y_hat = pd.DataFrame({"prediction_set": y_hat})
    y_hat = convert_expected_effectiveness_nonconformity_input_types(y_hat)
    singleton_counts = [len(ps[0]) == 1 for ps in y_hat]
    num_singletons = sum(singleton_counts)
    return num_singletons / y_hat.shape[0]


def avg_c(y_hat: Union[np.ndarray, pd.DataFrame, pd.Series]) -> float:
    """
    Calculate the average number of labels in all prediction sets.
    :param y_hat: Predicted probabilities of shape (n_samples, 1) where each row is a set of classes.
    """
    if isinstance(y_hat, np.ndarray):
        y_hat = pd.DataFrame({"prediction_set": y_hat})
    y_hat = convert_expected_effectiveness_nonconformity_input_types(y_hat)
    set_counts = [len(ps[0]) for ps in y_hat]
    return np.mean(np.asarray(set_counts))


def prediction_interval_spans(
    prediction_intervals: pd.DataFrame, alphas
) -> Dict[float, float]:
    """
    Calculate the mean span or width prediction intervals.

    This checks the distance between low and high band for each alpha.
    :param prediction_intervals: Predicted bands according to provided confidence levels.
    :param alphas: List of alphas indicating which confidence levels to check
    """
    interval_spans = {}
    for alpha in alphas:
        interval_spans[alpha] = np.mean(
            (
                prediction_intervals[f"{1-alpha}_high"]
                - prediction_intervals[f"{alpha}_low"]
            )
        )

    return interval_spans



from typing import Any, Callable, List

import numpy as np
import pandas as pd

from bluecast.conformal_prediction.base_classes import (
    ConformalPredictionWrapperBaseClass,
)
from bluecast.conformal_prediction.nonconformity_measures import hinge_loss


class ConformalPredictionWrapper(ConformalPredictionWrapperBaseClass):
    """
    Calibrate a model instance given a calibration set.

    :param model: An already fitted model instance of any type
    :param nonconformity_measure_scorer: A function object to calculate nonconformity scores with args
        y_calibration, preds
    """

    def __init__(self, model: Any, nonconformity_measure_scorer: Callable = hinge_loss):
        self.model = model
        self.nonconformity_measure_scorer = nonconformity_measure_scorer
        self.nonconformity_scores: List[float] = []

    def calibrate(self, x_calibration: pd.DataFrame, y_calibration: pd.Series):
        """
        Calibrate a model instance given a calibration set.

        :param x_calibration: Calibration set features. Must be unseen data for the model
        :param y_calibration: Calibration set labels or values
        """
        preds = self.model.predict_proba(x_calibration)
        self.nonconformity_scores = self.nonconformity_measure_scorer(
            y_calibration, preds
        )
        return self.nonconformity_scores

    def predict(self, x):
        return self.model.predict(x)

    def predict_proba(self, x):
        return self.model.predict_proba(x)

    def predict_interval(self, x: pd.DataFrame) -> np.ndarray:
        preds = self.model.predict_proba(x)
        if (
            len(preds.shape) == 1
        ):  # if a binary classifier only gives proba of target class
            preds = np.asarray([1 - preds, preds]).T

        if isinstance(preds, pd.DataFrame):
            preds = preds.values

        # calculate p-values (credibility intervals) for each label
        p_values = []
        # loop through all rows
        for _, pred in enumerate(preds):
            p_values_each_class = []
            # within each row loop through each class score
            for _class_idx, pred_j in enumerate(pred):
                p_values_each_class.append(
                    (
                        np.sum(
                            self.nonconformity_scores
                            >= self.nonconformity_measure_scorer(
                                np.asarray([1]), np.asarray([pred_j])
                            )
                        )
                        + 1
                    )
                    / (len(self.nonconformity_scores) + 1)
                )
            p_values.append(p_values_each_class)

        return np.asarray(p_values)

    def predict_sets(self, x: pd.DataFrame, alpha: float = 0.05) -> List[set[int]]:
        credible_intervals = self.predict_interval(x)
        prediction_sets = []
        for row in credible_intervals:
            prediction_set = []
            for class_idx, credible_interval in enumerate(row):
                if credible_interval >= alpha:
                    prediction_set.append(class_idx)
            prediction_sets.append(set(prediction_set))
        return prediction_sets



import numpy as np
from sklearn.datasets import make_regression
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import train_test_split

from bluecast.conformal_prediction.conformal_prediction_regression import (
    ConformalPredictionRegressionWrapper,
)


def test_calibrate():
    # Generate some random data
    X, y = make_regression(n_samples=100, n_features=5, noise=0.1, random_state=42)
    X_train, X_calibrate, y_train, y_calibrate = train_test_split(
        X, y, test_size=0.2, random_state=42
    )

    # Train a linear regression model
    model = LinearRegression()
    model.fit(X_train, y_train)

    # Create a ConformalPredictionRegressionWrapper instance
    wrapper = ConformalPredictionRegressionWrapper(model)

    # Calibrate the wrapper
    wrapper.calibrate(X_calibrate, y_calibrate)

    # Check that the nonconformity scores have been calculated
    assert isinstance(wrapper.nonconformity_scores, np.ndarray)
    assert len(wrapper.nonconformity_scores) == len(y_calibrate)

    # Check that the nonconformity scores are floats
    for score in wrapper.nonconformity_scores:
        assert isinstance(score, float)


def test_predict():
    # Generate some random data
    X, y = make_regression(n_samples=100, n_features=5, noise=0.1, random_state=42)
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.2, random_state=42
    )

    # Train a linear regression model
    model = LinearRegression()
    model.fit(X_train, y_train)

    # Create a ConformalPredictionRegressionWrapper instance
    wrapper = ConformalPredictionRegressionWrapper(model)

    # Make predictions
    y_pred = wrapper.predict(X_test)

    # Check that the predictions have the correct shape
    assert y_pred.shape == (len(X_test),)


def test_predict_interval():
    # Generate some random data
    X, y = make_regression(n_samples=100, n_features=5, noise=0.1, random_state=42)
    X_train, X_calibrate, y_train, y_calibrate = train_test_split(
        X, y, test_size=0.2, random_state=42
    )

    # Train a linear regression model
    model = LinearRegression()
    model.fit(X_train, y_train)

    # Create a ConformalPredictionRegressionWrapper instance
    wrapper = ConformalPredictionRegressionWrapper(model)

    # Calibrate the wrapper
    wrapper.calibrate(X_calibrate, y_calibrate)

    # Make predictions
    alphas = [0.05, 0.01]
    y_pred_interval = wrapper.predict_interval(X_calibrate, alphas)

    # Check that the predictions have the correct shape
    assert y_pred_interval.shape == (len(X_calibrate), len(alphas) * 2)


def test_predict_interval_with_single_data_point():
    # Generate some random data
    X, y = make_regression(n_samples=100, n_features=5, noise=0.1, random_state=42)
    X_train, _, y_train, _ = train_test_split(X, y, test_size=0.2, random_state=42)

    # Train a linear regression model
    model = LinearRegression()
    model.fit(X_train, y_train)

    # Create a ConformalPredictionRegressionWrapper instance
    wrapper = ConformalPredictionRegressionWrapper(model)

    # Calibrate the wrapper with a single data point calibration set
    wrapper.calibrate(X_train, y_train)

    # Make predictions
    alphas = [0.05, 0.01]
    y_pred_interval = wrapper.predict_interval(X_train, alphas)

    # Check that the predictions have the correct shape
    assert y_pred_interval.shape == (len(X_train), len(alphas) * 2)


def test_predict_interval_with_single_quantile():
    # Generate some random data
    X, y = make_regression(n_samples=100, n_features=5, noise=0.1, random_state=42)
    X_train, X_calibrate, y_train, y_calibrate = train_test_split(
        X, y, test_size=0.2, random_state=42
    )

    # Train a linear regression model
    model = LinearRegression()
    model.fit(X_train, y_train)

    # Create a ConformalPredictionRegressionWrapper instance
    wrapper = ConformalPredictionRegressionWrapper(model)

    # Calibrate the wrapper
    wrapper.calibrate(X_calibrate, y_calibrate)

    # Make predictions
    alphas = [0.05]
    y_pred_interval = wrapper.predict_interval(X_calibrate, alphas)

    # Check that the predictions have the correct shape
    assert y_pred_interval.shape == (len(X_calibrate), len(alphas) * 2)
    assert np.sum(y_pred_interval.iloc[:, 0] <= y_pred_interval.iloc[:, -1]) == len(
        X_calibrate
    )

    # Make predictions
    alphas = [0.05, 0.1, 0.5]
    y_pred_interval = wrapper.predict_interval(X_calibrate, alphas)
    assert y_pred_interval.shape == (len(X_calibrate), len(alphas) * 2)



import numpy as np
import pandas as pd
import pytest
from xgboost import XGBClassifier

from bluecast.preprocessing.feature_selection import BoostARoota, BoostaRootaWrapper


def test_boostaroota_wrapper_binary():
    df = pd.DataFrame(
        {
            "feature1": [i for i in range(100)],
            "feature2": [i for i in range(100, 200)],
        }
    )
    targets = pd.Series([0 for _i in range(50)] + [1 for _i in range(50)])
    # Initialize the wrapper
    boosta = BoostaRootaWrapper(class_problem="binary", random_state=45)
    # Fit the model
    trans_df, trans_targets = boosta.fit_transform(df, targets)

    assert trans_targets.equals(targets)


def test_boostaroota_wrapper_multiclass():
    df = pd.DataFrame(
        {
            "feature1": [i for i in range(300)],
            "feature2": [i for i in range(100, 400)],
        }
    )
    targets = pd.Series(
        [0 for _i in range(100)] + [1 for _i in range(100)] + [2 for _i in range(100)]
    )
    # Initialize the wrapper
    boosta = BoostaRootaWrapper(class_problem="multiclass", random_state=45)
    # Fit the model
    trans_df, trans_targets = boosta.fit_transform(df, targets)

    assert trans_targets.equals(targets)


def test_boostaroota_wrapper_regression():
    df = pd.DataFrame(
        {
            "feature1": [i for i in range(200)],
            "feature2": [i for i in range(100, 300)],
        }
    )
    targets = pd.Series([0 for _i in range(100)] + [1 for _i in range(100)])
    # Initialize the wrapper
    boosta = BoostaRootaWrapper(class_problem="regression", random_state=45)
    # Fit the model
    trans_df, trans_targets = boosta.fit_transform(df, targets)

    assert trans_targets.equals(targets)


@pytest.fixture
def synthetic_data():
    np.random.seed(0)
    df = pd.DataFrame(
        np.random.randn(100, 10), columns=[f"feature_{i}" for i in range(10)]
    )
    targets_binary = pd.Series(np.random.randint(0, 2, size=100))
    targets_multiclass = pd.Series(np.random.randint(0, 3, size=100))
    targets_regression = pd.Series(np.random.randn(100))
    return df, targets_binary, targets_multiclass, targets_regression


def test_boostaroota_wrapper_initialization():
    with pytest.raises(ValueError):
        _ = BoostaRootaWrapper(class_problem="unknown", random_state=42)


def test_boostaroota_wrapper_fit_transform_binary(synthetic_data):
    df, targets_binary, _, _ = synthetic_data
    wrapper = BoostaRootaWrapper(class_problem="binary", random_state=42)
    df_transformed, targets_transformed = wrapper.fit_transform(df, targets_binary)
    assert not df_transformed.empty


def test_boostaroota_wrapper_fit_transform_multiclass(synthetic_data):
    df, _, targets_multiclass, _ = synthetic_data
    wrapper = BoostaRootaWrapper(class_problem="multiclass", random_state=42)
    df_transformed, targets_transformed = wrapper.fit_transform(df, targets_multiclass)
    assert not df_transformed.empty


def test_boostaroota_wrapper_fit_transform_regression(synthetic_data):
    df, _, _, targets_regression = synthetic_data
    wrapper = BoostaRootaWrapper(class_problem="regression", random_state=42)
    df_transformed, targets_transformed = wrapper.fit_transform(df, targets_regression)
    assert not df_transformed.empty


def test_boostaroota_wrapper_transform(synthetic_data):
    df, targets_binary, _, _ = synthetic_data
    wrapper = BoostaRootaWrapper(class_problem="binary", random_state=42)
    wrapper.fit_transform(df, targets_binary)
    df_transformed, _ = wrapper.transform(df)
    assert not df_transformed.empty


def test_boostaroota_initialization_errors():
    with pytest.raises(ValueError):
        _ = BoostARoota(metric=None, clf=None)

    with pytest.raises(ValueError):
        _ = BoostARoota(metric="mlogloss", clf=None, cutoff=-1)

    with pytest.raises(ValueError):
        _ = BoostARoota(metric="mlogloss", clf=None, iters=0)

    with pytest.raises(ValueError):
        _ = BoostARoota(metric="mlogloss", clf=None, delta=1.5)


def test_boostaroota_warnings():
    with pytest.warns(UserWarning):
        _ = BoostARoota(metric="mlogloss", clf=XGBClassifier())

    with pytest.warns(UserWarning):
        _ = BoostARoota(metric="mlogloss", clf=None, delta=0.01)

    with pytest.warns(UserWarning):
        _ = BoostARoota(metric="mlogloss", clf=None, max_rounds=0)


def test_boostaroota_fit_transform(synthetic_data):
    df, targets_binary, _, _ = synthetic_data
    clf = XGBClassifier()
    br = BoostARoota(clf=clf)
    br.fit(df, targets_binary)
    transformed_df = br.transform(df)
    assert not transformed_df.empty

    transformed_df_fit_transform = br.fit_transform(df, targets_binary)
    assert not transformed_df_fit_transform.empty


def test_boostaroota_transform_without_fit(synthetic_data):
    df, _, _, _ = synthetic_data
    br = BoostARoota(clf=XGBClassifier())
    with pytest.raises(ValueError):
        _ = br.transform(df)



"""Xgboost classification model.

This module contains a wrapper for the Xgboost classification model. It can be used to train and/or tune the model.
It also calculates class weights for imbalanced datasets. The weights may or may not be used deepending on the
hyperparameter tuning.
"""

import logging
from copy import deepcopy
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import optuna
import pandas as pd
import xgboost as xgb
from sklearn.model_selection import StratifiedKFold
from sklearn.utils import class_weight

from bluecast.config.training_config import (
    TrainingConfig,
    XgboostFinalParamConfig,
    XgboostTuneParamsConfig,
)
from bluecast.evaluation.eval_metrics import ClassificationEvalWrapper
from bluecast.experimentation.tracking import ExperimentTracker
from bluecast.ml_modelling.base_classes import BaseClassMlModel
from bluecast.ml_modelling.parameter_tuning_utils import (
    get_params_based_on_device,
    sample_data,
    update_hyperparam_space_after_nth_trial,
    update_params_based_on_tree_method,
    update_params_with_best_params,
)
from bluecast.preprocessing.custom import CustomPreprocessing


class XgboostModel(BaseClassMlModel):
    """Train and/or tune Xgboost classification model."""

    def __init__(
        self,
        class_problem: Literal["binary", "multiclass"],
        conf_training: Optional[TrainingConfig] = None,
        conf_xgboost: Optional[XgboostTuneParamsConfig] = None,
        conf_params_xgboost: Optional[XgboostFinalParamConfig] = None,
        experiment_tracker: Optional[ExperimentTracker] = None,
        custom_in_fold_preprocessor: Optional[CustomPreprocessing] = None,
        cat_columns: Optional[List[Union[str, float, int]]] = None,
        single_fold_eval_metric_func: Optional[ClassificationEvalWrapper] = None,
    ):
        self.model: Optional[xgb.XGBClassifier] = None
        self.class_problem = class_problem
        self.conf_training = conf_training
        self.conf_xgboost = conf_xgboost
        self.conf_params_xgboost = conf_params_xgboost
        self.experiment_tracker = experiment_tracker
        self.custom_in_fold_preprocessor = custom_in_fold_preprocessor
        self.single_fold_eval_metric_func = single_fold_eval_metric_func
        if self.conf_training:
            self.random_generator = np.random.default_rng(
                self.conf_training.global_random_state
            )
        else:
            self.random_generator = np.random.default_rng(0)
        self.cat_columns = cat_columns
        self.best_score: float = np.inf

        if not self.single_fold_eval_metric_func:
            self.single_fold_eval_metric_func = ClassificationEvalWrapper()

    def calculate_class_weights(self, y: pd.Series) -> Dict[str, float]:
        """Calculate class weights of target column."""
        classes_weights = class_weight.compute_sample_weight(
            class_weight="balanced", y=y
        )
        return classes_weights

    def check_load_confs(self):
        """Load multiple configs or load default configs instead."""
        logging.info("Start loading existing or default config files..")
        if not self.conf_training:
            self.conf_training = TrainingConfig()
            logging.info("Load default TrainingConfig.")
        else:
            logging.info("Found provided TrainingConfig.")

        if not self.conf_xgboost:
            self.conf_xgboost = XgboostTuneParamsConfig()
            logging.info("Load default XgboostTuneParamsConfig.")
        else:
            logging.info("Found provided XgboostTuneParamsConfig.")

        if not self.conf_params_xgboost:
            self.conf_params_xgboost = XgboostFinalParamConfig()
            logging.info("Load default XgboostFinalParamConfig.")
        else:
            logging.info("Found provided XgboostFinalParamConfig.")

    def fit(
        self,
        x_train: pd.DataFrame,
        x_test: pd.DataFrame,
        y_train: pd.Series,
        y_test: pd.Series,
    ) -> xgb.Booster:
        """Train Xgboost model. Includes hyperparameter tuning on default."""
        logging.info("Start fitting Xgboost model.")
        self.check_load_confs()

        if (
            not self.conf_params_xgboost
            or not self.conf_training
            or not self.experiment_tracker
        ):
            raise ValueError(
                "conf_params_xgboost, conf_training or experiment_tracker is None"
            )

        if not self.conf_training.show_detailed_tuning_logs:
            optuna.logging.set_verbosity(optuna.logging.WARNING)

        if self.conf_training.autotune_model:
            self.autotune(x_train, x_test, y_train, y_test)
            print("Finished hyperparameter tuning")

        if self.conf_training.enable_grid_search_fine_tuning:
            self.fine_tune(x_train, x_test, y_train, y_test)
            print("Finished Grid search fine tuning")

        logging.info("Start final model training")
        if self.custom_in_fold_preprocessor:
            x_train, y_train = self.custom_in_fold_preprocessor.fit_transform(
                x_train, y_train
            )
            x_test, y_test = self.custom_in_fold_preprocessor.transform(
                x_test, y_test, predicton_mode=False
            )

        if self.conf_training.use_full_data_for_final_model:
            logging.info(
                """Union train and test data for final model training based on TrainingConfig
             param 'use_full_data_for_final_model'"""
            )
            x_train = pd.concat([x_train, x_test]).reset_index(drop=True)
            y_train = pd.concat([y_train, y_test]).reset_index(drop=True)

            if self.cat_columns and self.conf_training.cat_encoding_via_ml_algorithm:
                x_train[self.cat_columns] = x_train[self.cat_columns].astype("category")

        d_train, d_test = self.create_d_matrices(x_train, y_train, x_test, y_test)
        eval_set = [(d_test, "test")]

        steps = self.conf_params_xgboost.params.pop("steps", 300)

        if self.conf_training.early_stopping_rounds and self.conf_xgboost:
            early_stop = xgb.callback.EarlyStopping(
                rounds=self.conf_training.early_stopping_rounds,
                metric_name=self.conf_xgboost.xgboost_eval_metric,
                data_name="test",
                save_best=self.conf_params_xgboost.params["booster"] != "gblinear",
            )
            callbacks = [early_stop]
        else:
            callbacks = None

        if self.conf_training.hypertuning_cv_folds == 1 and self.conf_xgboost:
            self.model = xgb.train(
                self.conf_params_xgboost.params,
                d_train,
                num_boost_round=steps,
                evals=eval_set,
                verbose_eval=self.conf_xgboost.verbosity_during_final_model_training,
                callbacks=callbacks,
            )
        elif self.conf_xgboost:
            self.model = xgb.train(
                self.conf_params_xgboost.params,
                d_train,
                num_boost_round=steps,
                early_stopping_rounds=self.conf_training.early_stopping_rounds,
                evals=eval_set,
                verbose_eval=self.conf_xgboost.verbosity_during_final_model_training,
                callbacks=callbacks,
            )
        logging.info("Finished training")
        return self.model

    def autotune(
        self,
        x_train: pd.DataFrame,
        x_test: pd.DataFrame,
        y_train: pd.Series,
        y_test: pd.Series,
    ) -> None:
        """Tune hyperparameters.

        An alternative config can be provided to overwrite the hyperparameter search space.
        """
        logging.info("Start hyperparameter tuning of Xgboost model.")
        if (
            not self.conf_params_xgboost
            or not self.conf_training
            or not self.experiment_tracker
        ):
            raise ValueError(
                "conf_params_xgboost, conf_training or experiment_tracker is None"
            )

        self.check_load_confs()

        if (
            not self.conf_params_xgboost
            or not self.conf_training
            or not self.conf_xgboost
            or not self.experiment_tracker
        ):
            raise ValueError(
                "At least one of the configs or experiment_tracker is None, which is not allowed"
            )

        train_on = get_params_based_on_device(
            self.conf_training, self.conf_params_xgboost, self.conf_xgboost
        )

        x_train, x_test, y_train, y_test = sample_data(
            x_train, x_test, y_train, y_test, self.conf_training
        )

        def objective(trial):
            self.conf_xgboost = update_hyperparam_space_after_nth_trial(
                trial,
                self.conf_xgboost,
                self.conf_training.update_hyperparameter_search_space_after_nth_trial,
            )

            param = {
                "validate_parameters": False,
                "objective": self.conf_xgboost.xgboost_objective,
                "booster": self.conf_xgboost.booster,
                "eval_metric": self.conf_xgboost.xgboost_eval_metric,
                "num_class": y_train.nunique(),
                "eta": trial.suggest_float(
                    "eta",
                    self.conf_xgboost.eta_min,
                    self.conf_xgboost.eta_max,
                    log=True,
                ),
                "max_depth": trial.suggest_int(
                    "max_depth",
                    self.conf_xgboost.max_depth_min,
                    self.conf_xgboost.max_depth_max,
                    log=False,
                ),
                "alpha": trial.suggest_float(
                    "alpha",
                    self.conf_xgboost.alpha_min,
                    self.conf_xgboost.alpha_max,
                    log=True,
                ),
                "lambda": trial.suggest_float(
                    "lambda",
                    self.conf_xgboost.lambda_min,
                    self.conf_xgboost.lambda_max,
                    log=True,
                ),
                "gamma": trial.suggest_float(
                    "gamma",
                    self.conf_xgboost.gamma_min,
                    self.conf_xgboost.gamma_max,
                    log=True,
                ),
                "min_child_weight": trial.suggest_float(
                    "min_child_weight",
                    self.conf_xgboost.min_child_weight_min,
                    self.conf_xgboost.min_child_weight_max,
                    log=True,
                ),
                "subsample": trial.suggest_float(
                    "subsample",
                    self.conf_xgboost.sub_sample_min,
                    self.conf_xgboost.sub_sample_max,
                ),
                "colsample_bytree": trial.suggest_float(
                    "colsample_bytree",
                    self.conf_xgboost.col_sample_by_tree_min,
                    self.conf_xgboost.col_sample_by_tree_max,
                ),
                "colsample_bylevel": trial.suggest_float(
                    "colsample_bylevel",
                    self.conf_xgboost.col_sample_by_level_min,
                    self.conf_xgboost.col_sample_by_level_max,
                ),
                "steps": trial.suggest_int(
                    "steps",
                    self.conf_xgboost.steps_min,
                    self.conf_xgboost.steps_max,
                    log=True,
                ),
            }
            params = {**param, **train_on}
            sample_weight = trial.suggest_categorical("sample_weight", [True, False])
            params = update_params_based_on_tree_method(
                params, trial, self.conf_xgboost
            )

            if sample_weight:
                classes_weights = self.calculate_class_weights(y_train)
                d_train = xgb.DMatrix(
                    x_train,
                    label=y_train,
                    weight=classes_weights,
                    enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
                )
            else:
                d_train = xgb.DMatrix(
                    x_train,
                    label=y_train,
                    enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
                )

            d_test = xgb.DMatrix(
                x_test,
                label=y_test,
                enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
            )

            pruning_callback = optuna.integration.XGBoostPruningCallback(
                trial, f"test-{self.conf_xgboost.xgboost_eval_metric}"
            )

            steps = params.pop("steps", 300)

            if self.conf_training.hypertuning_cv_folds == 1:
                try:
                    return self.train_single_fold_model(
                        d_train, d_test, y_test, params, steps, pruning_callback
                    )
                except Exception as e:
                    logging.error(f"Error during training: {e}. Pruning trial")
                    trial.should_prune()
            elif (
                self.conf_training.hypertuning_cv_folds > 1
                and self.conf_training.precise_cv_tuning
            ):

                return self._fine_tune_precise(params, x_train, y_train, x_test, y_test)
            else:
                skf = StratifiedKFold(
                    n_splits=5,
                    random_state=self.conf_training.global_random_state,
                    shuffle=self.conf_training.shuffle_during_training,
                )
                folds = []
                for train_index, test_index in skf.split(x_train, y_train.tolist()):
                    folds.append((train_index.tolist(), test_index.tolist()))

                result = xgb.cv(
                    params=params,
                    dtrain=d_train,
                    num_boost_round=steps,
                    # early_stopping_rounds=self.conf_training.early_stopping_rounds,  # not recommended as per docs: https://xgboost.readthedocs.io/en/stable/python/sklearn_estimator.html
                    nfold=self.conf_training.hypertuning_cv_folds,
                    as_pandas=True,
                    seed=self.conf_training.global_random_state,
                    callbacks=[pruning_callback],
                    shuffle=self.conf_training.shuffle_during_training,
                    # stratified=True,
                    folds=folds,
                )

                adjusted_score = result[
                    f"test-{self.conf_xgboost.xgboost_eval_metric}-mean"
                ].values[-1]

                # track results
                if len(self.experiment_tracker.experiment_id) == 0:
                    new_id = 0
                else:
                    new_id = self.experiment_tracker.experiment_id[-1] + 1
                self.experiment_tracker.add_results(
                    experiment_id=new_id,
                    score_category="cv_score",
                    training_config=self.conf_training,
                    model_parameters=params,
                    eval_scores=adjusted_score,
                    metric_used="adjusted ml logloss",
                    metric_higher_is_better=False,
                )

                return adjusted_score

        for rst in range(self.conf_training.autotune_n_random_seeds):
            logging.info(
                f"Hyperparameter tuning using random seed {self.conf_training.global_random_state + rst}"
            )
            sampler = optuna.samplers.TPESampler(
                multivariate=True,
                seed=self.conf_training.global_random_state + rst,
                n_startup_trials=self.conf_training.optuna_sampler_n_startup_trials,
                warn_independent_sampling=False,
            )
            study = optuna.create_study(
                direction=self.conf_xgboost.xgboost_eval_metric_tune_direction,
                sampler=sampler,
                study_name="xgboost tuning",
                pruner=optuna.pruners.MedianPruner(
                    n_startup_trials=20,
                    n_warmup_steps=20,
                ),
            )

            study.optimize(
                objective,
                n_trials=self.conf_training.hyperparameter_tuning_rounds,
                timeout=self.conf_training.hyperparameter_tuning_max_runtime_secs,
                gc_after_trial=True,
                show_progress_bar=True,
            )

            if self.conf_training.plot_hyperparameter_tuning_overview:
                try:
                    fig = optuna.visualization.plot_optimization_history(study)
                    fig.show()
                    fig = optuna.visualization.plot_param_importances(
                        study  # , evaluator=FanovaImportanceEvaluator()
                    )
                    fig.show()
                except (ZeroDivisionError, RuntimeError, ValueError):
                    pass

            if study.best_value < self.best_score:
                self.best_score = study.best_value
                logging.info(
                    f"New best score: {study.best_value} from random seed  {self.conf_training.global_random_state + rst}"
                )
                xgboost_best_param = study.best_trial.params

                self.conf_params_xgboost.params = {
                    "validate_parameters": False,
                    "objective": self.conf_xgboost.xgboost_objective,  # OR  'binary:logistic' #the loss function being used
                    "booster": self.conf_xgboost.booster,
                    "eval_metric": self.conf_xgboost.xgboost_eval_metric,
                    "num_class": y_train.nunique(),
                    "max_depth": xgboost_best_param[
                        "max_depth"
                    ],  # maximum depth of the decision trees being trained
                    "alpha": xgboost_best_param["alpha"],
                    "lambda": xgboost_best_param["lambda"],
                    "gamma": xgboost_best_param["gamma"],
                    "min_child_weight": xgboost_best_param["min_child_weight"],
                    "subsample": xgboost_best_param["subsample"],
                    "colsample_bytree": xgboost_best_param["colsample_bytree"],
                    "colsample_bylevel": xgboost_best_param["colsample_bylevel"],
                    "eta": xgboost_best_param["eta"],
                    "steps": xgboost_best_param["steps"],
                }
                self.conf_params_xgboost.params = {
                    **self.conf_params_xgboost.params,
                    **train_on,
                }
                self.conf_params_xgboost.params = update_params_with_best_params(
                    self.conf_params_xgboost.params, xgboost_best_param
                )
                logging.info(f"Best params: {self.conf_params_xgboost.params}")
                print(f"Best params: {self.conf_params_xgboost.params}")
                self.conf_params_xgboost.sample_weight = xgboost_best_param[
                    "sample_weight"
                ]

    def create_d_matrices(self, x_train, y_train, x_test, y_test):
        if self.conf_params_xgboost.sample_weight:
            classes_weights = self.calculate_class_weights(y_train)
            d_train = xgb.DMatrix(
                x_train,
                label=y_train,
                weight=classes_weights,
                enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
            )
        else:
            d_train = xgb.DMatrix(
                x_train,
                label=y_train,
                enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
            )
        d_test = xgb.DMatrix(
            x_test,
            label=y_test,
            enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
        )
        return d_train, d_test

    def train_single_fold_model(
        self, d_train, d_test, y_test, param, steps, pruning_callback
    ):
        eval_set = [(d_test, "test")]
        if self.conf_training.early_stopping_rounds and self.conf_xgboost:
            early_stop = xgb.callback.EarlyStopping(
                rounds=self.conf_training.early_stopping_rounds,
                metric_name=self.conf_xgboost.xgboost_eval_metric,
                data_name="test",
                save_best=param["booster"] != "gblinear",
            )
            callbacks = [early_stop]
        else:
            callbacks = None

        model = xgb.train(
            param,
            d_train,
            num_boost_round=steps,
            evals=eval_set,
            callbacks=callbacks,
            verbose_eval=self.conf_xgboost.verbosity_during_hyperparameter_tuning,
        )
        preds = model.predict(d_test)
        matthew = self.single_fold_eval_metric_func.classification_eval_func_wrapper(
            y_test, preds
        )

        # track results
        if len(self.experiment_tracker.experiment_id) == 0:
            new_id = 0
        else:
            new_id = self.experiment_tracker.experiment_id[-1] + 1
        self.experiment_tracker.add_results(
            experiment_id=new_id,
            score_category="simple_train_test_score",
            training_config=self.conf_training,
            model_parameters=param,
            eval_scores=matthew,
            metric_used="matthew_inverse",
            metric_higher_is_better=False,
        )
        return matthew

    def _fine_tune_precise(
        self,
        tuned_params: Dict[str, Any],
        x_train: pd.DataFrame,
        y_train: pd.Series,
        x_test: pd.DataFrame,
        y_test: pd.Series,
    ):
        steps = tuned_params.pop("steps", 300)

        if not self.conf_training:
            self.conf_training = TrainingConfig()
            logging.info(
                "Could not find Training config. Falling back to default values"
            )

        stratifier = StratifiedKFold(
            n_splits=self.conf_training.hypertuning_cv_folds,
            shuffle=self.conf_training.shuffle_during_training,
            random_state=self.conf_training.global_random_state,
        )

        fold_losses = []
        for _fn, (trn_idx, val_idx) in enumerate(
            stratifier.split(x_train, y_train.astype(int))
        ):
            X_train_fold, X_val_fold = (
                x_train.iloc[trn_idx],
                x_train.iloc[val_idx],
            )
            y_train_fold, y_val_fold = (
                y_train.iloc[trn_idx],
                y_train.iloc[val_idx],
            )
            if self.custom_in_fold_preprocessor:
                (
                    X_train_fold,
                    y_train_fold,
                ) = self.custom_in_fold_preprocessor.fit_transform(
                    X_train_fold, y_train_fold
                )
                (
                    X_test_fold,
                    y_test_fold,
                ) = self.custom_in_fold_preprocessor.transform(x_test, y_test)
                (
                    X_val_fold,
                    y_val_fold,
                ) = self.custom_in_fold_preprocessor.transform(
                    X_val_fold, y_val_fold, predicton_mode=False
                )
            else:
                X_test_fold, y_test_fold = x_test, y_test

            d_test = xgb.DMatrix(
                X_val_fold,
                label=y_val_fold,
                enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
            )

            if not self.conf_params_xgboost:
                self.conf_params_xgboost = XgboostFinalParamConfig()
                logging.info(
                    "Could not find XgboostFinalParamConfig. Falling back to default settings."
                )

            if self.conf_params_xgboost.sample_weight:
                classes_weights = self.calculate_class_weights(y_train_fold)
                d_train = xgb.DMatrix(
                    X_train_fold,
                    label=y_train_fold,
                    weight=classes_weights,
                    enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
                )
            else:
                d_train = xgb.DMatrix(
                    X_train_fold,
                    label=y_train_fold,
                    enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
                )
            eval_set = [(d_test, "test")]

            if not self.conf_xgboost:
                self.conf_xgboost = XgboostTuneParamsConfig()
                logging.info(
                    "Could not find XgboostTuneParamsConfig. Falling back to defaults."
                )

            model = xgb.train(
                tuned_params,
                d_train,
                num_boost_round=steps,
                evals=eval_set,
                verbose_eval=self.conf_xgboost.verbosity_during_hyperparameter_tuning,
            )
            d_eval = xgb.DMatrix(
                X_test_fold,
                label=y_test_fold,
                enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
            )
            preds = model.predict(d_eval)

            if self.single_fold_eval_metric_func:
                loss = (
                    self.single_fold_eval_metric_func.classification_eval_func_wrapper(
                        y_test_fold, preds
                    )
                )
            else:
                raise ValueError("No single_fold_eval_metric_func could be found")
            fold_losses.append(loss)

        matthews_mean = np.mean(np.asarray(fold_losses))

        if self.experiment_tracker and self.conf_training:
            # track results
            if len(self.experiment_tracker.experiment_id) == 0:
                new_id = 0
            else:
                new_id = self.experiment_tracker.experiment_id[-1] + 1
            self.experiment_tracker.add_results(
                experiment_id=new_id,
                score_category="oof_score",
                training_config=self.conf_training,
                model_parameters=tuned_params,
                eval_scores=matthews_mean,
                metric_used="matthew_inverse",
                metric_higher_is_better=False,
            )
        return matthews_mean

    def fine_tune(
        self,
        x_train: pd.DataFrame,
        x_test: pd.DataFrame,
        y_train: pd.Series,
        y_test: pd.Series,
    ) -> None:
        logging.info("Start grid search fine tuning of Xgboost model.")
        if (
            not self.conf_params_xgboost
            or not self.conf_training
            or not self.conf_xgboost
            or not self.experiment_tracker
        ):
            raise ValueError(
                "At least one of the configs or experiment_tracker is None, which is not allowed"
            )

        def objective(trial):
            d_train, d_test = self.create_d_matrices(x_train, y_train, x_test, y_test)

            pruning_callback = optuna.integration.XGBoostPruningCallback(
                trial, f"test-{self.conf_xgboost.xgboost_eval_metric}"
            )
            # copy best params to not overwrite them
            tuned_params = deepcopy(self.conf_params_xgboost.params)
            min_child_weight_space = trial.suggest_float(
                "min_child_weight",
                self.conf_params_xgboost.params["min_child_weight"] * 0.9,
                self.conf_params_xgboost.params["min_child_weight"] * 1.1,
                log=False,
            )
            lambda_space = trial.suggest_float(
                "lambda",
                self.conf_params_xgboost.params["lambda"] * 0.9,
                self.conf_params_xgboost.params["lambda"] * 1.1,
                log=False,
            )
            gamma_space = trial.suggest_float(
                "gamma",
                self.conf_params_xgboost.params["gamma"] * 0.9,
                self.conf_params_xgboost.params["gamma"] * 1.1,
                log=False,
            )
            eta_space = trial.suggest_float(
                "eta",
                self.conf_params_xgboost.params["eta"] * 0.9,
                self.conf_params_xgboost.params["eta"] * 1.1,
                log=False,
            )

            tuned_params["lambda"] = lambda_space
            tuned_params["min_child_weight"] = min_child_weight_space
            tuned_params["gamma"] = gamma_space
            tuned_params["eta"] = eta_space

            steps = tuned_params.pop("steps", 300)

            if self.conf_training.hypertuning_cv_folds == 1:
                try:
                    return self.train_single_fold_model(
                        d_train, d_test, y_test, tuned_params, steps, pruning_callback
                    )
                except Exception as e:
                    logging.error(f"Error during training: {e}. Pruning trial")
                    trial.should_prune()
            elif (
                self.conf_training.hypertuning_cv_folds > 1
                and self.conf_training.precise_cv_tuning
            ):
                return self._fine_tune_precise(
                    tuned_params,
                    x_train,
                    y_train,
                    x_test,
                    y_test,
                )
            else:
                skf = StratifiedKFold(
                    n_splits=5,
                    random_state=self.conf_training.global_random_state,
                    shuffle=self.conf_training.shuffle_during_training,
                )
                folds = []
                for train_index, test_index in skf.split(x_train, y_train):
                    folds.append((train_index.tolist(), test_index.tolist()))

                result = xgb.cv(
                    params=tuned_params,
                    dtrain=d_train,
                    num_boost_round=steps,
                    # early_stopping_rounds=self.conf_training.early_stopping_rounds,  # not recommended as per docs: https://xgboost.readthedocs.io/en/stable/python/sklearn_estimator.html
                    nfold=self.conf_training.hypertuning_cv_folds,
                    as_pandas=True,
                    seed=self.conf_training.global_random_state,
                    callbacks=[pruning_callback],
                    shuffle=self.conf_training.shuffle_during_training,
                    # stratified=True,
                    folds=folds,
                )

                adjusted_score = result[
                    f"test-{self.conf_xgboost.xgboost_eval_metric}-mean"
                ].values[-1]

                # track results
                if len(self.experiment_tracker.experiment_id) == 0:
                    new_id = 0
                else:
                    new_id = self.experiment_tracker.experiment_id[-1] + 1
                self.experiment_tracker.add_results(
                    experiment_id=new_id,
                    score_category="cv_score",
                    training_config=self.conf_training,
                    model_parameters=tuned_params,
                    eval_scores=adjusted_score,
                    metric_used="adjusted ml logloss",
                    metric_higher_is_better=False,
                )

                return adjusted_score

        self.check_load_confs()
        if (
            isinstance(self.conf_params_xgboost.params["min_child_weight"], float)
            and isinstance(self.conf_params_xgboost.params["lambda"], float)
            and isinstance(self.conf_params_xgboost.params["gamma"], float)
            and isinstance(self.conf_params_xgboost.params["eta"], float)
        ):
            search_space = {
                "min_child_weight": np.linspace(
                    self.conf_params_xgboost.params["min_child_weight"] * 0.9,
                    self.conf_params_xgboost.params["min_child_weight"] * 1.1,
                    self.conf_training.gridsearch_nb_parameters_per_grid,
                    dtype=float,
                ),
                "lambda": np.linspace(
                    self.conf_params_xgboost.params["lambda"] * 0.9,
                    self.conf_params_xgboost.params["lambda"] * 1.1,
                    self.conf_training.gridsearch_nb_parameters_per_grid,
                    dtype=float,
                ),
                "gamma": np.linspace(
                    self.conf_params_xgboost.params["gamma"] * 0.9,
                    self.conf_params_xgboost.params["gamma"] * 1.1,
                    self.conf_training.gridsearch_nb_parameters_per_grid,
                    dtype=float,
                ),
                "eta": np.linspace(
                    self.conf_params_xgboost.params["eta"] * 0.9,
                    self.conf_params_xgboost.params["eta"] * 1.1,
                    self.conf_training.gridsearch_nb_parameters_per_grid,
                    dtype=float,
                ),
            }
        else:
            ValueError("Some parameters are not floats or strings")

        study = optuna.create_study(
            direction=self.conf_xgboost.xgboost_eval_metric_tune_direction,
            sampler=optuna.samplers.GridSampler(search_space),
            pruner=optuna.pruners.MedianPruner(n_startup_trials=20, n_warmup_steps=50),
        )
        study.optimize(
            objective,
            n_trials=self.conf_training.gridsearch_nb_parameters_per_grid
            ** len(search_space.keys()),
            timeout=self.conf_training.gridsearch_tuning_max_runtime_secs,
            gc_after_trial=True,
            show_progress_bar=True,
        )

        if self.conf_training.plot_hyperparameter_tuning_overview:
            try:
                fig = optuna.visualization.plot_optimization_history(study)
                fig.show()
                fig = optuna.visualization.plot_param_importances(
                    study  # , evaluator=FanovaImportanceEvaluator()
                )
                fig.show()
            except (ZeroDivisionError, RuntimeError, ValueError):
                pass

        best_score_cv = self.best_score

        if study.best_value < self.best_score or not self.conf_training.autotune_model:
            self.best_score = study.best_value
            xgboost_grid_best_param = study.best_trial.params
            self.conf_params_xgboost.params["min_child_weight"] = (
                xgboost_grid_best_param["min_child_weight"]
            )
            self.conf_params_xgboost.params["lambda"] = xgboost_grid_best_param[
                "lambda"
            ]
            self.conf_params_xgboost.params["gamma"] = xgboost_grid_best_param["gamma"]
            self.conf_params_xgboost.params["eta"] = xgboost_grid_best_param["eta"]
            logging.info(
                f"Grid search improved eval metric from {best_score_cv} to {self.best_score}."
            )
            logging.info(f"Best params: {self.conf_params_xgboost.params}")
            print(f"Best params: {self.conf_params_xgboost.params}")
        else:
            logging.info(
                f"Grid search could not improve eval metric of {best_score_cv}. Best score reached was {study.best_value}"
            )

    def predict(self, df: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
        """Predict on unseen data."""
        logging.info("Start predicting on new data using Xgboost model.")
        if not self.conf_xgboost or not self.conf_training:
            raise ValueError("conf_params_xgboost or conf_training is None")

        if self.custom_in_fold_preprocessor:
            df, _ = self.custom_in_fold_preprocessor.transform(
                df, None, predicton_mode=True
            )

        d_test = xgb.DMatrix(
            df,
            enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
        )

        if not self.model:
            raise Exception("No trained model has been found.")

        if not self.conf_params_xgboost:
            raise Exception("No model configuration file has been found.")

        partial_probs = self.model.predict(d_test)
        if self.class_problem == "binary":
            predicted_probs = np.asarray([line[1] for line in partial_probs])
            predicted_classes = (
                predicted_probs > self.conf_params_xgboost.classification_threshold
            )
        else:
            predicted_probs = partial_probs
            predicted_classes = np.asarray([np.argmax(line) for line in partial_probs])
        logging.info("Finished predicting")
        return predicted_probs, predicted_classes

    def predict_proba(self, df: pd.DataFrame) -> np.ndarray:
        """Predict class scores on unseen data."""
        logging.info("Start predicting on new data using Xgboost model.")
        if not self.conf_xgboost or not self.conf_training:
            raise ValueError("conf_params_xgboost or conf_training is None")

        if self.custom_in_fold_preprocessor:
            df, _ = self.custom_in_fold_preprocessor.transform(
                df, None, predicton_mode=True
            )

        d_test = xgb.DMatrix(
            df,
            enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
        )

        if not self.model:
            raise Exception("No trained model has been found.")

        if not self.conf_params_xgboost:
            raise Exception("No model configuration file has been found.")

        partial_probs = self.model.predict(d_test)
        if self.class_problem == "binary":
            predicted_probs = np.asarray([line[1] for line in partial_probs])
        else:
            predicted_probs = partial_probs
        logging.info("Finished predicting")
        return predicted_probs



import logging
from typing import Any, List, Literal, Optional, Tuple, Union

import numpy as np
import pandas as pd
from sklearn.model_selection import RepeatedStratifiedKFold

from bluecast.blueprints.cast import BlueCast
from bluecast.config.training_config import (
    TrainingConfig,
    XgboostFinalParamConfig,
    XgboostTuneParamsConfig,
)
from bluecast.conformal_prediction.conformal_prediction import (
    ConformalPredictionWrapper,
)
from bluecast.evaluation.eval_metrics import ClassificationEvalWrapper
from bluecast.experimentation.tracking import ExperimentTracker
from bluecast.ml_modelling.xgboost import XgboostModel
from bluecast.preprocessing.custom import CustomPreprocessing
from bluecast.preprocessing.feature_selection import BoostaRootaWrapper


class BlueCastCV:
    """Wrapper to train and predict multiple BlueCast instances.

    A custom splitter can be provided.

    :param :class_problem: Takes a string containing the class problem type. Either "binary" or "multiclass".
    :param :target_column: Takes a string containing the name of the target column.
    :param :cat_columns: Takes a list of strings containing the names of the categorical columns. If not provided,
        BlueCast will infer these automatically.
    :param :date_columns: Takes a list of strings containing the names of the date columns. If not provided,
        BlueCast will infer these automatically.
    :param :time_split_column: Takes a string containing the name of the time split column. If not provided,
        BlueCast will not split the data by time or order, but do a random split instead.
    :param :ml_model: Takes an instance of a XgboostModel class. If not provided, BlueCast will instantiate one.
        This is an API to pass any model class. Inherit the baseclass from ml_modelling.base_model.BaseModel.
    :param custom_in_fold_preprocessor: Takes an instance of a CustomPreprocessing class. Allows users to eeecute
        preprocessing after the train test split within cv folds. This will be executed only if precise_cv_tuning in
        the conf_Training is True. Custom ML models need to implement this themselves. This step is only useful when
        the proprocessing step has a high chance of overfitting otherwise (i.e: oversampling techniques).
    :param custom_preprocessor: Takes an instance of a CustomPreprocessing class. Allows users to inject custom
        preprocessing steps which take place right after the train test spit.
    :param custom_last_mile_computation: Takes an instance of a CustomPreprocessing class. Allows users to inject custom
        preprocessing steps which take place right before the model training.
    :param experiment_tracker: Takes an instance of an ExperimentTracker class. If not provided this will be initialized
        automatically.
    :param single_fold_eval_metric_func: Takes a function which calculates the evaluation metric for a single fold.
           Default is matthews_corrcoef. This function is used to calculate the evaluation metric for each fold during
           hyperparameter tuning when hyperparameter_tuning_rounds = 1 (default). Lower must be better.
    """

    def __init__(
        self,
        class_problem: Literal["binary", "multiclass"] = "binary",
        cat_columns: Optional[List[Union[str, float, int]]] = None,
        stratifier: Optional[Any] = None,
        conf_training: Optional[TrainingConfig] = None,
        conf_xgboost: Optional[XgboostTuneParamsConfig] = None,
        conf_params_xgboost: Optional[XgboostFinalParamConfig] = None,
        experiment_tracker: Optional[ExperimentTracker] = None,
        custom_in_fold_preprocessor: Optional[CustomPreprocessing] = None,
        custom_last_mile_computation: Optional[CustomPreprocessing] = None,
        custom_preprocessor: Optional[CustomPreprocessing] = None,
        custom_feature_selector: Optional[
            Union[BoostaRootaWrapper, CustomPreprocessing]
        ] = None,
        ml_model: Optional[Union[XgboostModel, Any]] = None,
        single_fold_eval_metric_func: Optional[ClassificationEvalWrapper] = None,
    ):
        self.class_problem = class_problem
        self.conf_xgboost = conf_xgboost
        self.conf_training = conf_training
        self.conf_params_xgboost = conf_params_xgboost
        self.custom_in_fold_preprocessor = custom_in_fold_preprocessor
        self.custom_preprocessor = custom_preprocessor
        self.custom_feature_selector = custom_feature_selector
        self.custom_last_mile_computation = custom_last_mile_computation
        self.bluecast_models: List[BlueCast] = []
        self.stratifier = stratifier
        self.ml_model = ml_model
        self.single_fold_eval_metric_func = single_fold_eval_metric_func
        self.conformal_prediction_wrapper: Optional[ConformalPredictionWrapper] = None

        if not cat_columns:
            self.cat_columns = []
        else:
            self.cat_columns = cat_columns

        if experiment_tracker:
            self.experiment_tracker = experiment_tracker
        else:
            self.experiment_tracker = ExperimentTracker()

        if not self.conf_params_xgboost:
            self.conf_params_xgboost = XgboostFinalParamConfig()

        if not self.conf_training:
            self.conf_training = TrainingConfig()

        if not self.conf_xgboost:
            self.conf_xgboost = XgboostTuneParamsConfig()

        if not self.single_fold_eval_metric_func:
            self.single_fold_eval_metric_func = ClassificationEvalWrapper()

    def prepare_data(
        self, df: pd.DataFrame, target: str
    ) -> Tuple[pd.DataFrame, pd.Series]:
        df = df.reset_index(drop=True)
        y = df[target]
        X = df.drop(target, axis=1)
        return X, y

    def show_oof_scores(self, metric: str = "matthews") -> Tuple[float, float]:
        """
        Show out of fold scores.

        When calling BlueCastCVRegression's fit_eval function multiple BlueCastRegression
        instances are called and each of them predicts on unseen/oof data.

        This function collects these scores and return mean and average of them.

        :param metric: String indicating which metric shall be returned.
        :return: Tuple with (mean, std) of oof scores
        """
        all_metrics = []
        for bluecast_instance in self.bluecast_models:
            if bluecast_instance.eval_metrics:
                score = bluecast_instance.eval_metrics.get(metric)
                all_metrics.append(score)

        score_mean = np.asarray(all_metrics).mean()
        score_std = np.asarray(all_metrics).std()
        logging.info(
            f"The mean out of fold {metric} score is {score_mean} with an std of {score_std}"
        )

        return score_mean, score_std

    def fit(self, df: pd.DataFrame, target_col: str) -> None:
        """Fit multiple BlueCast instances on different data splits.

        Input df is expected the target column."""
        X, y = self.prepare_data(df, target_col)

        if not self.conf_training:
            self.conf_training = TrainingConfig()

        if not self.stratifier:
            self.stratifier = RepeatedStratifiedKFold(
                n_splits=self.conf_training.bluecast_cv_train_n_model[0],
                n_repeats=self.conf_training.bluecast_cv_train_n_model[1],
                random_state=self.conf_training.global_random_state,
            )

        for fn, (trn_idx, val_idx) in enumerate(self.stratifier.split(X, y)):
            X_train, X_val = X.iloc[trn_idx], X.iloc[val_idx]
            y_train, y_val = y.iloc[trn_idx], y.iloc[val_idx]
            x_train = pd.concat([X_train, X_val], ignore_index=True)
            y_train = pd.concat([y_train, y_val], ignore_index=True)

            X_train = x_train.reset_index(drop=True)
            y_train = y_train.reset_index(drop=True)
            X_train[target_col] = y_train.values

            self.conf_training.global_random_state += (
                self.conf_training.increase_random_state_in_bluecast_cv_by
            )
            logging.info(
                f"Start fitting model number {fn} with random seed {self.conf_training.global_random_state}"
            )

            automl = BlueCast(
                class_problem=self.class_problem,
                cat_columns=self.cat_columns,
                conf_training=self.conf_training,
                conf_xgboost=self.conf_xgboost,
                conf_params_xgboost=self.conf_params_xgboost,
                experiment_tracker=self.experiment_tracker,
                custom_in_fold_preprocessor=self.custom_in_fold_preprocessor,
                custom_preprocessor=self.custom_preprocessor,
                custom_feature_selector=self.custom_feature_selector,
                custom_last_mile_computation=self.custom_last_mile_computation,
                ml_model=self.ml_model,
                single_fold_eval_metric_func=self.single_fold_eval_metric_func,
            )
            automl.fit(X_train, target_col=target_col)
            self.bluecast_models.append(automl)

            # overwrite experiment tracker to pass it into next iteration
            self.experiment_tracker = automl.experiment_tracker

    def fit_eval(self, df: pd.DataFrame, target_col: str) -> Tuple[float, float]:
        """Fit multiple BlueCast instances on different data splits.

        Input df is expected the target column. Evaluation is executed on out-of-fold dataset.
        in each split.
        :param df: Pandas DataFrame that includes the target column
        :param target_col: String indicating the name of the target column
        :returns Tuple of (oof_mean, oof_std) with scores on unseen data during eval
        """
        X, y = self.prepare_data(df, target_col)

        if not self.conf_training:
            self.conf_training = TrainingConfig()

        if not self.stratifier:
            self.stratifier = RepeatedStratifiedKFold(
                n_splits=self.conf_training.bluecast_cv_train_n_model[0],
                n_repeats=self.conf_training.bluecast_cv_train_n_model[1],
                random_state=self.conf_training.global_random_state,
            )

        for fn, (trn_idx, val_idx) in enumerate(self.stratifier.split(X, y)):
            X_train, X_val = X.iloc[trn_idx], X.iloc[val_idx]
            y_train, y_val = y.iloc[trn_idx], y.iloc[val_idx]

            X_train.loc[:, target_col] = y_train

            self.conf_training.global_random_state += (
                self.conf_training.increase_random_state_in_bluecast_cv_by
            )
            logging.info(
                f"Start fitting model number {fn} with random seed {self.conf_training.global_random_state}"
            )

            automl = BlueCast(
                class_problem=self.class_problem,
                cat_columns=self.cat_columns,
                conf_training=self.conf_training,
                conf_xgboost=self.conf_xgboost,
                conf_params_xgboost=self.conf_params_xgboost,
                experiment_tracker=self.experiment_tracker,
                custom_in_fold_preprocessor=self.custom_in_fold_preprocessor,
                custom_preprocessor=self.custom_preprocessor,
                custom_feature_selector=self.custom_feature_selector,
                custom_last_mile_computation=self.custom_last_mile_computation,
                ml_model=self.ml_model,
                single_fold_eval_metric_func=self.single_fold_eval_metric_func,
            )
            automl.fit_eval(X_train, X_val, y_val, target_col=target_col)
            self.bluecast_models.append(automl)

            # overwrite experiment tracker to pass it into next iteration
            self.experiment_tracker = automl.experiment_tracker

        oof_mean, oof_std = self.show_oof_scores()
        return oof_mean, oof_std

    def predict(
        self,
        df: pd.DataFrame,
        return_sub_models_preds: bool = False,
        save_shap_values: bool = False,
    ) -> Tuple[Union[pd.DataFrame, pd.Series], Union[pd.DataFrame, pd.Series]]:
        """Predict on unseen data using multiple trained BlueCast instances.

        :param df: Pandas DataFrame with unseen data
        :param return_sub_models_preds: If true will return a DataFrame with the predictions of each model for each class
            stored in separate columns.
        :param save_shap_values: If True, calculates and saves shap values, so they can be used to plot
            waterfall plots for selected rows o demand.
        """
        result_df = pd.DataFrame()  # Create an empty DataFrame to store the results
        or_cols = df.columns
        prob_cols: list[str] = []
        class_cols: list[str] = []
        for fn, pipeline in enumerate(self.bluecast_models):
            y_probs, y_classes = pipeline.predict(
                df.loc[:, or_cols], save_shap_values=save_shap_values
            )
            if self.class_problem == "multiclass":
                proba_cols = [
                    f"class_{col}_proba_model_{fn}" for col in range(y_probs.shape[1])
                ]
                result_df[proba_cols] = y_probs
                result_df[f"classes_{fn}"] = y_classes
                for col in proba_cols:
                    prob_cols.append(col)
                class_cols.append(f"classes_{fn}")

            else:
                result_df[f"proba_{fn}"] = y_probs
                result_df[f"classes_{fn}"] = y_classes
                prob_cols.append(f"proba_{fn}")
                class_cols.append(f"classes_{fn}")

        if self.class_problem == "multiclass":
            if return_sub_models_preds:
                return result_df.loc[:, prob_cols], result_df.loc[:, class_cols]
            else:
                classes = result_df.loc[:, class_cols].mode(axis=1)[0].astype(int)

                if self.bluecast_models[0].feat_type_detector:
                    if (
                        self.bluecast_models[0].target_label_encoder
                        and self.bluecast_models[0].feat_type_detector
                    ):
                        classes = self.bluecast_models[
                            0
                        ].target_label_encoder.label_encoder_reverse_transform(classes)

                mean_class_proba_cols = []
                for col_idx in range(y_probs.shape[1]):
                    class_proba_cols = [
                        f"class_{col_idx}_proba_model_{fn}"
                        for fn, pipeline in enumerate(self.bluecast_models)
                    ]
                    result_df[f"mean_proba_class_{col_idx}"] = result_df.loc[
                        :, class_proba_cols
                    ].mean(axis=1)
                    mean_class_proba_cols.append(f"mean_proba_class_{col_idx}")

                return (
                    result_df.loc[:, mean_class_proba_cols],
                    classes,
                )
        else:
            if return_sub_models_preds:
                return result_df.loc[:, prob_cols], result_df.loc[:, class_cols]
            else:
                return (
                    result_df.loc[:, prob_cols].mean(axis=1),
                    result_df.loc[:, prob_cols].mean(axis=1) > 0.5,
                )

    def predict_proba(
        self,
        df: pd.DataFrame,
        return_sub_models_preds: bool = False,
        save_shap_values: bool = False,
    ) -> Union[pd.DataFrame, pd.Series]:
        """Predict on unseen data using multiple trained BlueCast instances.

        :param df: Pandas DataFrame with unseen data
        :param return_sub_models_preds: If true will return a DataFrame with the predictions of each model for each class
            stored in separate columns.
        :param save_shap_values: If True, calculates and saves shap values, so they can be used to plot
            waterfall plots for selected rows o demand.
        """
        result_df = pd.DataFrame()  # Create an empty DataFrame for storing results
        or_cols = df.columns
        prob_cols: list[str] = []
        for fn, pipeline in enumerate(self.bluecast_models):
            y_probs, _y_classes = pipeline.predict(
                df.loc[:, or_cols], save_shap_values=save_shap_values
            )
            if self.class_problem == "multiclass":
                proba_cols = [
                    f"class_{col}_proba_model_{fn}" for col in range(y_probs.shape[1])
                ]
                result_df[proba_cols] = y_probs
                for col in proba_cols:
                    prob_cols.append(col)

            else:
                result_df[f"proba_{fn}"] = y_probs
                prob_cols.append(f"proba_{fn}")

        if self.class_problem == "multiclass":
            if return_sub_models_preds:
                return result_df.loc[:, prob_cols]
            else:
                # TODO: Take mean by class instead of overall
                mean_class_proba_cols = []
                for col_idx in range(y_probs.shape[1]):
                    class_proba_cols = [
                        f"class_{col_idx}_proba_model_{fn}"
                        for fn, pipeline in enumerate(self.bluecast_models)
                    ]
                    result_df[f"mean_proba_class_{col_idx}"] = result_df.loc[
                        :, class_proba_cols
                    ].mean(axis=1)
                    mean_class_proba_cols.append(f"mean_proba_class_{col_idx}")
                return result_df.loc[:, mean_class_proba_cols]
        else:
            if return_sub_models_preds:
                return result_df.loc[:, prob_cols]
            else:
                return result_df.loc[:, prob_cols].mean(axis=1)

    def calibrate(
        self, x_calibration: pd.DataFrame, y_calibration: pd.Series, **kwargs
    ) -> None:
        """Calibrate the model.

        Via this function the nonconformity measures are taken and used to predict calibrated sets via the
        predict_sets function, or to return p-values of a row for being the class via the predict_p_values function.
        This calibrates the blended prediction of all sub models.
        :param: x_calibration: Pandas DataFrame without target column, that has not been seen by the model during
            training.
        :param y_calibration: Pandas Series holding the target value, hat has not been seen by the model during
            training.
        """
        if self.bluecast_models[0].target_label_encoder:
            x_calibration[self.bluecast_models[0].target_column] = y_calibration
            x_calibration = self.bluecast_models[
                0
            ].target_label_encoder.transform_target_labels(
                x_calibration, self.bluecast_models[0].target_column
            )
            y_calibration = x_calibration.pop(self.bluecast_models[0].target_column)

        self.conformal_prediction_wrapper = ConformalPredictionWrapper(self, **kwargs)
        self.conformal_prediction_wrapper.calibrate(x_calibration, y_calibration)

    def predict_p_values(self, df: pd.DataFrame) -> np.ndarray:
        """Create p-values for each class.

        The p_values indicate the probability of being the correct class.
        :param df: Pandas DataFrame holding unseen data
        :returns: Numpy array where each column holds p-values of a row being the class. If string labels were passed
            each column maps the index of target_label_encoder.target_label_mapping stored in this class.
        """
        if self.conformal_prediction_wrapper:
            pred_interval = self.conformal_prediction_wrapper.predict_interval(df)
            return pred_interval
        else:
            raise ValueError(
                """This instance has not been calibrated yet. Make use of calibrate to fit the
            ConformalPredictionWrapper."""
            )

    def predict_sets(self, df: pd.DataFrame, alpha: float = 0.05) -> pd.DataFrame:
        """Create prediction sets based on a certain confidence level.

        Conformal prediction guarantees, that the correct label is present in the prediction sets with a probability of
        1 - alpha.
        :param df: Pandas DataFrame holding unseen data
        :param alpha: Float indicating the desired confidence level.
        :returns a Pandas DataFrame with a column called 'prediction_set' holding a nested set with predicted classes.
        """
        if self.conformal_prediction_wrapper:
            pred_sets = self.conformal_prediction_wrapper.predict_sets(df, alpha)
            # transform numerical values back to original strings for the end user
            if self.bluecast_models[0].target_label_encoder:
                reverse_mapping = {
                    value: key
                    for key, value in self.bluecast_models[
                        0
                    ].target_label_encoder.target_label_mapping.items()
                }
                string_pred_sets = []
                for numerical_set in pred_sets:
                    # Convert numerical labels to string labels
                    string_set = {reverse_mapping[label] for label in numerical_set}
                    string_pred_sets.append(string_set)
                return pd.DataFrame({"prediction_set": string_pred_sets})
            else:
                return pd.DataFrame({"prediction_set": pred_sets})
        else:
            raise ValueError(
                """This instance has not been calibrated yet. Make use of calibrate to fit the
            ConformalPredictionWrapper."""
            )



from typing import Tuple

import numpy as np
import pandas as pd

from bluecast.conformal_prediction.nonconformity_measures import (
    brier_score,
    hinge_loss,
    margin_nonconformity_measure,
)


def create_synthetic_binary_data() -> Tuple[np.ndarray, pd.Series]:
    y_true_binary = pd.Series([0, 1, 0])
    synthetic_results_binary = np.asarray([0.7, 0.3, 0.0])
    return synthetic_results_binary, y_true_binary


def create_synthetic_multiclass_data() -> Tuple[np.ndarray, pd.Series]:
    y_true_multiclass = pd.Series([0, 1, 2])
    synthetic_results_multiclass = np.asarray(
        [[0.7, 0.3, 0.0], [0.7, 0.3, 0.0], [0.7, 0.3, 0.0]]
    )
    return synthetic_results_multiclass, y_true_multiclass


def test_hinge_loss():
    synthetic_results_binary, y_true_binary = create_synthetic_binary_data()
    synthetic_results_multiclass, y_true_multiclass = create_synthetic_multiclass_data()
    assert np.allclose(
        hinge_loss(y_true_binary, synthetic_results_binary), np.asarray([0.7, 0.7, 0.0])
    )
    assert np.allclose(
        hinge_loss(y_true_multiclass, synthetic_results_multiclass),
        np.asarray([0.3, 0.7, 1.0]),
    )


def test_margin_nonconformity_measure():
    synthetic_results_binary, y_true_binary = create_synthetic_binary_data()
    synthetic_results_multiclass, y_true_multiclass = create_synthetic_multiclass_data()
    result_binary = margin_nonconformity_measure(
        y_true_binary, synthetic_results_binary
    )
    result_multiclass = margin_nonconformity_measure(
        y_true_multiclass, synthetic_results_multiclass
    )
    assert np.allclose(result_binary, np.asarray([-0.4, -0.4, 1.0]))
    assert np.allclose(result_multiclass, np.asarray([0.4, -0.4, -0.7]))


def test_brier_score():
    synthetic_results_binary, y_true_binary = create_synthetic_binary_data()
    synthetic_results_multiclass, y_true_multiclass = create_synthetic_multiclass_data()
    assert np.allclose(
        brier_score(y_true_binary, synthetic_results_binary),
        np.asarray([0.49, 0.49, 0.0]),
    )
    assert np.allclose(
        brier_score(y_true_multiclass, synthetic_results_multiclass),
        np.asarray([0.09, 0.49, 1.0]),
    )



from typing import Tuple, Union

import numpy as np
import pandas as pd


def convert_to_numpy(
    y_true: pd.Series, y_hat: Union[np.ndarray, pd.Series, pd.DataFrame]
) -> Tuple[np.ndarray, np.ndarray]:
    if isinstance(y_hat, pd.Series):
        y_hat = y_hat.values
    elif isinstance(y_hat, pd.DataFrame):  # pd..core.frame.DataFrame
        y_hat = y_hat.values
    else:
        pass

    if isinstance(y_true, pd.Series):
        y_true = y_true.values
    elif isinstance(y_true, pd.DataFrame):
        y_true = y_true.values
    else:
        pass
    return y_true, y_hat


def hinge_loss(
    y_true: pd.Series, y_hat: Union[np.ndarray, pd.Series, pd.DataFrame]
) -> np.ndarray:
    """
    Calculate Hinge loss per row.

    To compute the nonconformity score, take the probability score of the true class and
    subtract it from 1.
    :param y_true: True labels
    :param y_hat: Predicted probabilities
    :return: Hinge loss per row
    """
    hinge_losses = []
    if len(y_hat.shape) == 1:  # if a binary classifier only gives proba of target class
        y_hat = np.asarray([1 - y_hat, y_hat]).T

    y_true, y_hat = convert_to_numpy(y_true, y_hat)

    for true_class, preds_arr in zip(y_true, y_hat):
        hinge_losses.append(1 - preds_arr[true_class])
    return np.asarray(hinge_losses)


def margin_nonconformity_measure(
    y_true: pd.Series, y_hat: Union[np.ndarray, pd.Series, pd.DataFrame]
) -> np.ndarray:
    """
    Calculate margin nonconformity score per row.

    The margin nonconformity measure is defined as the difference between the predicted probability of
    the most likely incorrect class label and the predicted probability of the true label.
    :param y_true: True labels
    :param y_hat: Predicted probabilities
    :return: Margin nonconformity score loss per row
    """
    mnm_losses = []
    if len(y_hat.shape) == 1:  # if a binary classifier only gives proba of target class
        y_hat = np.asarray([1 - y_hat, y_hat]).T

    y_true, y_hat = convert_to_numpy(y_true, y_hat)

    for true_class, preds_arr in zip(y_true, y_hat):
        prob_y_true = preds_arr[true_class]

        # remove true label proba
        probas_y_false = np.delete(preds_arr, true_class)
        prob_y_most_likely_false = np.max(probas_y_false)
        mnm_losses.append(prob_y_true - prob_y_most_likely_false)
    return np.asarray(mnm_losses)


def brier_score(
    y_true: pd.Series, y_hat: Union[np.ndarray, pd.Series, pd.DataFrame]
) -> np.ndarray:
    """
    Calculate Brier score per row.

    It calculates the squared difference between the predicted probabilities and the actual binary results.
    The score’s values can range from 0 (perfect accuracy) to 1 (complete inaccuracy).
    :param y_true: True labels
    :param y_hat: Predicted probabilities
    :return: Brier score loss per row
    """
    if len(y_hat.shape) == 1:  # if a binary classifier only gives proba of target class
        y_hat = np.asarray([1 - y_hat, y_hat]).T

    y_true, y_hat = convert_to_numpy(y_true, y_hat)

    brier_losses = []
    for true_class, preds_arr in zip(y_true, y_hat):
        brier_losses.append((1 - preds_arr[true_class]) ** 2)
    return np.asarray(brier_losses)



from unittest.mock import patch

import pytest

from bluecast.config.training_config import (
    TrainingConfig,
    XgboostFinalParamConfig,
    XgboostTuneParamsConfig,
)
from bluecast.ml_modelling.xgboost import XgboostModel


@pytest.fixture
def default_model():
    return XgboostModel(class_problem="binary")


def test_check_load_confs_defaults(default_model):
    with patch("bluecast.ml_modelling.xgboost.logging") as mock_logger:
        default_model.check_load_confs()

        # Check if the defaults are loaded
        assert isinstance(default_model.conf_training, TrainingConfig)
        assert isinstance(default_model.conf_xgboost, XgboostTuneParamsConfig)
        assert isinstance(default_model.conf_params_xgboost, XgboostFinalParamConfig)

        # Ensure logger was called with default config loading messages
        assert mock_logger.call_count == 0


def test_check_load_confs_partial():
    conf_training = TrainingConfig()
    model = XgboostModel(class_problem="binary", conf_training=conf_training)

    with patch("bluecast.ml_modelling.xgboost.logging") as mock_logger:
        model.check_load_confs()

        # Check if the provided config is used and defaults are loaded for the rest
        assert model.conf_training is conf_training
        assert isinstance(model.conf_xgboost, XgboostTuneParamsConfig)
        assert isinstance(model.conf_params_xgboost, XgboostFinalParamConfig)

        # Ensure logger was called with appropriate messages
        assert mock_logger.call_count == 0


def test_check_load_confs_all_provided():
    conf_training = TrainingConfig()
    conf_xgboost = XgboostTuneParamsConfig()
    conf_params_xgboost = XgboostFinalParamConfig()

    model = XgboostModel(
        class_problem="binary",
        conf_training=conf_training,
        conf_xgboost=conf_xgboost,
        conf_params_xgboost=conf_params_xgboost,
    )

    with patch("bluecast.ml_modelling.xgboost.logging") as mock_logger:
        model.check_load_confs()

        # Check if all provided configs are used
        assert model.conf_training is conf_training
        assert model.conf_xgboost is conf_xgboost
        assert model.conf_params_xgboost is conf_params_xgboost

        # Ensure logger was called with appropriate messages
        assert mock_logger.call_count == 0



"""Define training and common configuration parameters.

Pydantic dataclasses are used to define the configuration parameters. This allows for type checking and validation of
the configuration parameters. The configuration parameters are used in the training pipeline and in the evaluation
pipeline. Pydantic dataclasses are used to allow users a pythonic way to define the configuration parameters.
Default configurations can be loaded, adjusted and passed into the blueprints.
"""

from typing import Dict, List, Literal, Optional, Tuple

from pydantic import BaseModel
from pydantic.dataclasses import dataclass


class Config:
    arbitrary_types_allowed = True


class TrainingConfig(BaseModel):
    """Define general training parameters.

    :param global_random_state: Global random state to use for reproducibility.
    :param increase_random_state_in_bluecast_cv_by: In BlueCastCV multiple models are trained. Define by how much the
        random state changes with each additional model.
    :param shuffle_during_training: Whether to shuffle the data during training when hypertuning_cv_folds > 1.
    :param hyperparameter_tuning_rounds: Number of hyperparameter tuning rounds. Not used when custom ML model is passed.
    :param hyperparameter_tuning_max_runtime_secs: Maximum runtime in seconds for hyperparameter tuning. Not used when
        custom ML model is passed.
    :param hypertuning_cv_folds: Number of cross-validation folds to use for hyperparameter tuning. Not used when
        custom ML model is passed.
    :param sample_data_during_tuning: Whether to sample the data during tuning. Not used when custom ML model is passed.
    :param sample_data_during_tuning_alpha: Alpha value for sampling the data during tuning. The higher alpha the
        fewer samples will be left. Not used when custom ML model is passed.
    :param class_weight_during_dmatrix_creation: Whether to use class weights during DMatrix creation. Not used when
        custom ML model is passed.
    :param early_stopping_rounds: Number of early stopping rounds during final training or when hyperparameter tuning
        follows a single train-test split. Not used when custom ML model is passed.
    :param autotune_model: Whether to autotune the model. Not used when custom ML model is passed.
    :param autotune_on_device: Whether to autotune on CPU or GPU. Chose any of ["auto", "gpu", "cpu"].
        Not used when custom ML model is passed.
    :param autotune_n_random_seeds: Number of random seeds to use for autotuning. This changes Optuna's random seed only.
        Will be updated back after every nth trial back again. Not used when custom ML model is passed.
    :param update_hyperparameter_search_space_after_nth_trial: Update the hyperparameter search space after the nth trial.
        Not used when custom ML model is passed.
    :param plot_hyperparameter_tuning_overview: Whether to plot the hyperparameter tuning overview. Not used when custom
        ML model is passed.
    :param enable_feature_selection: Whether to enable recursive feature selection.
    :param calculate_shap_values: Whether to calculate shap values. Also used when custom ML model is passed. Not
        compatible with all ML models. See the SHAP documentation for more details.
    :param shap_waterfall_indices: List of sample indices to plot. Each index represents a sample (i.e.: [0, 1, 499]).
    :param show_dependence_plots_of_top_n_features: Maximum number of dependence plots to show. Not used when custom ML
        model is passed.
    :param store_shap_values_in_instance: Whether to store the SHAP values in the BlueCast instance. Not applicable when
        custom ML model is used.
    :param train_size: Train size to use for train-test split.
    :param train_split_stratify: Whether to stratify the train-test split. Not used when custom ML model is passed.
    :param use_full_data_for_final_model: Whether to use the full data for the final model. This might cause overfitting.
        Not used when custom ML model is passed.
    :param cardinality_threshold_for_onehot_encoding: Categorical features with a cardinality of less or equal
        this threshold will be onehot encoded. The rest will be target encoded. Will be ignored if
        cat_encoding_via_ml_algorithm is set to true.
    :param cat_encoding_via_ml_algorithm: Whether to use an ML algorithm for categorical encoding. If True, the
        categorical encoding is done via a ML algorithm. If False, the categorical encoding is done via a  target
        encoding in the preprocessing steps. See the ReadMe for more details.
    :param show_detailed_tuning_logs: Whether to show detailed tuning logs. Not used when custom ML model is passed.
    :param enable_grid_search_fine_tuning: After hyperparameter tuning run Gridsearch tuning on a fine-grained grid
        based on the previous hyperparameter tuning. Only possible when autotune_model is True.
    :param gridsearch_nb_parameters_per_grid: Decides how many steps the grid shall have per parameter.
    :param gridsearch_tuning_max_runtime_secs: Sets the maximum time in seconds the tuning shall run. This will finish
        the latest trial nd will exceed this limit though.
    :param experiment_name: Name of the experiment. Will be logged inside the ExperimentTracker.
    :param logging_file_path: Path to the logging file. If None, the logging will be printed to the Jupyter notebook
        instead.
    """

    global_random_state: int = 33
    increase_random_state_in_bluecast_cv_by: int = 200
    shuffle_during_training: bool = True
    hyperparameter_tuning_rounds: int = 200
    hyperparameter_tuning_max_runtime_secs: int = 3600
    hypertuning_cv_folds: int = 1
    sample_data_during_tuning: bool = False
    sample_data_during_tuning_alpha: float = 2.0
    precise_cv_tuning: bool = False
    early_stopping_rounds: Optional[int] = 10
    autotune_model: bool = True
    autotune_on_device: Literal["auto", "gpu", "cpu"] = "auto"
    autotune_n_random_seeds: int = 1
    update_hyperparameter_search_space_after_nth_trial: int = 25
    plot_hyperparameter_tuning_overview: bool = True
    enable_feature_selection: bool = False
    calculate_shap_values: bool = True
    shap_waterfall_indices: List[int] = [0]
    show_dependence_plots_of_top_n_features: int = 1
    store_shap_values_in_instance: bool = False
    train_size: float = 0.8
    train_split_stratify: bool = True
    use_full_data_for_final_model: bool = False
    cardinality_threshold_for_onehot_encoding: int = 5
    cat_encoding_via_ml_algorithm: bool = False
    show_detailed_tuning_logs: bool = False
    optuna_sampler_n_startup_trials: int = 10
    enable_grid_search_fine_tuning: bool = False
    gridsearch_tuning_max_runtime_secs: int = 3600
    gridsearch_nb_parameters_per_grid: int = 5
    bluecast_cv_train_n_model: Tuple[int, int] = (5, 1)
    logging_file_path: Optional[str] = None
    experiment_name: str = "new experiment"


class XgboostTuneParamsConfig(BaseModel):
    """Define hyperparameter tuning search space."""

    max_depth_min: int = 1
    max_depth_max: int = 10
    alpha_min: float = 1e-8
    alpha_max: float = 100
    lambda_min: float = 1
    lambda_max: float = 100
    gamma_min: float = 1e-8
    gamma_max: float = 100
    min_child_weight_min: float = 1
    min_child_weight_max: float = 100
    sub_sample_min: float = 0.5
    sub_sample_max: float = 1.0
    col_sample_by_tree_min: float = 0.5
    col_sample_by_tree_max: float = 1.0
    col_sample_by_level_min: float = 1.0
    col_sample_by_level_max: float = 1.0
    eta_min: float = 1e-3
    eta_max: float = 0.3
    steps_min: int = 50
    steps_max: int = 1000
    verbosity_during_hyperparameter_tuning: int = 0
    verbosity_during_final_model_training: int = 0
    booster: List[str] = ["gbtree"]  # "gblinear"
    grow_policy: List[str] = ["depthwise", "lossguide"]
    tree_method: List[str] = ["exact", "approx", "hist"]
    xgboost_objective: str = "multi:softprob"
    xgboost_eval_metric: str = "mlogloss"
    xgboost_eval_metric_tune_direction: Literal["minimize", "maximize"] = "minimize"


class XgboostTuneParamsRegressionConfig(BaseModel):
    """Define hyperparameter tuning search space."""

    max_depth_min: int = 1
    max_depth_max: int = 10
    alpha_min: float = 1e-8
    alpha_max: float = 100
    lambda_min: float = 1e-8
    lambda_max: float = 100
    gamma_min: float = 1e-8
    gamma_max: float = 100
    min_child_weight_min: float = 1
    min_child_weight_max: float = 100
    sub_sample_min: float = 0.5
    sub_sample_max: float = 1.0
    col_sample_by_tree_min: float = 0.5
    col_sample_by_tree_max: float = 1.0
    col_sample_by_level_min: float = 1.0
    col_sample_by_level_max: float = 1.0
    eta_min: float = 1e-3
    eta_max: float = 0.3
    steps_min: int = 50
    steps_max: int = 1000
    verbosity_during_hyperparameter_tuning: int = 0
    verbosity_during_final_model_training: int = 0
    booster: List[str] = ["gbtree"]  # "gblinear"
    grow_policy: List[str] = ["depthwise", "lossguide"]
    tree_method: List[str] = ["exact", "approx", "hist"]
    xgboost_objective: str = "reg:squarederror"
    xgboost_eval_metric: str = "rmse"
    xgboost_eval_metric_tune_direction: Literal["minimize", "maximize"] = "minimize"


@dataclass
class XgboostFinalParamConfig:
    """Define final hyper parameters."""

    params = {
        "booster": "gbtree",
        "max_depth": 10,  # maximum depth of the decision trees being trained
        "alpha": 0.0,
        "lambda": 1.0,
        "gamma": 0.0,
        "subsample": 1.0,
        "min_child_weight": 1.0,
        "colsample_bytree": 1.0,
        "colsample_bylevel": 1.0,
        "eta": 0.05,
        "steps": 1000,
        "objective": "multi:softprob",
        "eval_metric": "mlogloss",
        "tree_method": "hist",
        "grow_policy": "depthwise",
        "device": "cpu",
    }
    sample_weight: Optional[Dict[str, float]] = None
    classification_threshold: float = 0.5


@dataclass
class XgboostRegressionFinalParamConfig:
    """Define final hyper parameters."""

    params = {
        "booster": "gbtree",
        "max_depth": 10,  # maximum depth of the decision trees being trained
        "alpha": 0.0,
        "lambda": 1.0,
        "gamma": 0.0,
        "subsample": 1.0,
        "min_child_weight": 1.0,
        "colsample_bytree": 1.0,
        "colsample_bylevel": 1.0,
        "eta": 0.05,
        "steps": 1000,
        "objective": "reg:squarederror",
        "eval_metric": "rmse",
        "tree_method": "hist",
        "grow_policy": "depthwise",
        "device": "cpu",
    }



"""General utilities."""

import logging
import warnings
from typing import Any, Dict, Optional

import dill as pickle
import numpy as np
import xgboost as xgb

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger()


# Capture warnings and redirect them to logging
def warning_to_logger(message, category, filename, lineno, file=None, line=None):
    logger.warning(f"{filename}:{lineno}: {category.__name__}: {message}")


warnings.showwarning = warning_to_logger


def check_gpu_support() -> Dict[str, str]:
    logger.info("Start checking if GPU is available for usage.")
    data = np.random.rand(50, 2)
    label = np.random.randint(2, size=50)
    d_train = xgb.DMatrix(data, label=label)

    params_list = [
        {"device": "cuda", "tree_method": "gpu_hist"},
        {"device": "cuda"},
        {"tree_method": "gpu_hist"},
    ]

    for params in params_list:
        try:
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                xgb.train(params, d_train, num_boost_round=2)

                # Check if any captured warnings indicate GPU-related issues
                gpu_warning = any(
                    "GPU" in str(warn.message) or "gpu" in str(warn.message)
                    for warn in w
                )
                if gpu_warning:
                    logger.warning(f"GPU-related warning captured: {w}")

                # If no warnings or no GPU-related warnings, consider GPU support confirmed
                if not gpu_warning:
                    logger.info("Xgboost is using GPU with parameters: %s", params)
                    return params
                else:
                    logger.warning(
                        "GPU settings applied but GPU-related warning detected: %s",
                        params,
                    )
        except xgb.core.XGBoostError as e:
            logger.warning("Failed with params %s. Error: %s", params, str(e))

    # If no GPU parameters work, fall back to CPU
    params = {"tree_method": "hist", "device": "cpu"}
    logger.info("No GPU detected. Xgboost will use CPU with parameters: %s", params)
    return params


def save_to_production(
    class_instance: Any,
    file_path: Optional[str] = None,
    file_name: str = "automl_instance",
    file_type: str = ".dat",
) -> None:
    """
    Takes a pretrained model and saves it via dill.
    :param class_instance: Takes instance of a BlueCast class.
    :param file_path: Takes a string containing the full absolute path.
    :param file_name: Takes a string containing the whole file name.
    :param file_type: Takes the expected type of file to export.
    :return:
    """
    logging.info("Start saving class instance.")
    if file_path:
        full_path = file_path + file_name + file_type
    else:
        full_path = file_name + file_type
    filehandler = open(full_path, "wb")
    pickle.dump(class_instance, filehandler)
    filehandler.close()


def load_for_production(
    file_path: Optional[str] = None,
    file_name: str = "automl_instance",
    file_type: str = ".dat",
) -> Any:
    """
    Load in a pretrained auto ml model. This function will try to load the model as provided.
    It has a fallback logic to impute .dat as file_type in case the import fails initially.
    :param file_path: Takes a string containing the full absolute path.
    :param file_name: Takes a string containing the whole file name.
    :param file_type: Takes the expected type of file to import.
    :return: The loaded model object
    """
    logging.info("Start loading class instance.")
    if file_path:
        full_path = file_path + file_name
    else:
        full_path = file_name
    try:
        with open(full_path, "rb") as filehandler:
            automl_model = pickle.load(filehandler)
    except Exception:
        with open(full_path + file_type, "rb") as filehandler:
            automl_model = pickle.load(filehandler)
    return automl_model


def log_sampling(nb_rows: int, alpha: float = 2.0) -> int:
    """Return a number of samples.

    :param nb_rows: Number of rows in the dataset.
    :param alpha: Alpha value for sampling weight. Higher alpha values will result in less samples.
    """
    perc_reduction = np.log10(nb_rows) ** alpha / 100

    nb_samples = int(round(nb_rows * (1 - perc_reduction), 0))
    logging.info(f"Down sampling from {nb_rows} to {nb_samples} samples.")
    return nb_samples



import numpy as np
import pytest
from sklearn.metrics import (
    accuracy_score,
    log_loss,
    matthews_corrcoef,
    mean_squared_error,
)

from bluecast.evaluation.eval_metrics import (
    ClassificationEvalWrapper,
    RegressionEvalWrapper,
)


@pytest.fixture
def sample_data():
    y_true = np.array([0, 1, 1, 0])
    y_probs = np.array([[0.8, 0.2], [0.2, 0.8], [0.4, 0.6], [0.9, 0.1]])
    return y_true, y_probs


@pytest.fixture
def sample_data_regression():
    y_true = np.array([1.0, 5.0, 25.0])
    y_hat = np.array([3.0, 3.0, 28.0])
    return y_true, y_hat


def test_init():
    wrapper = ClassificationEvalWrapper()
    assert wrapper.higher_is_better
    assert wrapper.eval_against == "classes"
    assert wrapper.metric_func == matthews_corrcoef


def test_probas_all_classes(sample_data):
    y_true, y_probs = sample_data
    wrapper = ClassificationEvalWrapper(
        eval_against="probas_all_classes",
        metric_func=log_loss,
        higher_is_better=False,
        metric_name="logloss",
    )
    score = wrapper.classification_eval_func_wrapper(y_true, y_probs)
    expected_score = log_loss(y_true.tolist(), y_probs.tolist())
    assert score == expected_score


def test_probas_target_class(sample_data):
    y_true, y_probs = sample_data
    wrapper = ClassificationEvalWrapper(
        eval_against="probas_target_class", metric_func=log_loss, higher_is_better=False
    )
    score = wrapper.classification_eval_func_wrapper(y_true, y_probs)
    y_probs_best_class = np.asarray([line[1] for line in y_probs])
    expected_score = log_loss(y_true, y_probs_best_class)
    assert score == expected_score


def test_classes(sample_data):
    y_true, y_probs = sample_data
    wrapper = ClassificationEvalWrapper(
        eval_against="classes", metric_func=accuracy_score, higher_is_better=True
    )
    score = wrapper.classification_eval_func_wrapper(y_true, y_probs)
    y_classes = np.asarray([np.argmax(line) for line in y_probs])
    expected_score = accuracy_score(y_true, y_classes)
    assert score == -expected_score


def test_higher_is_better_false(sample_data):
    y_true, y_probs = sample_data
    wrapper = ClassificationEvalWrapper(
        higher_is_better=True, eval_against="classes", metric_func=accuracy_score
    )
    score = wrapper.classification_eval_func_wrapper(y_true, y_probs)
    y_classes = np.asarray([np.argmax(line) for line in y_probs])
    expected_score = accuracy_score(y_true, y_classes)
    assert score == -expected_score


def test_invalid_eval_against():
    with pytest.raises(ValueError):
        ClassificationEvalWrapper(eval_against="invalid_value")


def test_regression_init():
    wrapper = RegressionEvalWrapper()
    assert not wrapper.higher_is_better
    assert wrapper.metric_func == mean_squared_error


def test_regression_run_with_args(sample_data_regression):
    y_true, y_hat = sample_data_regression
    wrapper = RegressionEvalWrapper(
        higher_is_better=False, metric_func=mean_squared_error, **{"squared": False}
    )
    score = wrapper.regression_eval_func_wrapper(y_true, y_hat)
    expected_score = mean_squared_error(y_true, y_hat, squared=False)
    assert score == expected_score


def test_regression_run_witouth_args(sample_data_regression):
    y_true, y_hat = sample_data_regression
    wrapper = RegressionEvalWrapper(
        higher_is_better=False, metric_func=mean_squared_error
    )
    score = wrapper.regression_eval_func_wrapper(y_true, y_hat)
    expected_score = mean_squared_error(y_true, y_hat)
    assert score == expected_score



from typing import Dict, List, Union

import numpy as np
import pandas as pd

from bluecast.conformal_prediction.effectiveness_nonconformity_measures import (
    convert_expected_effectiveness_nonconformity_input_types,
)


def prediction_set_coverage(
    y_true: Union[np.ndarray, pd.Series],
    prediction_sets: Union[pd.Series, pd.DataFrame],
) -> float:
    """
    Calculate the percentyge of prediction sets that include the true label.

    This check can be used to validate that the model covers the true labels
    according to the alpha set during the prediction process.
    :param y_true: Ground truth labels.
    :param prediction_sets: Predicted probabilities of shape (n_samples, 1) where each row is a set of classes.
    """
    y_hat = convert_expected_effectiveness_nonconformity_input_types(prediction_sets)
    set_with_corr_label = [str(label) in str(ps) for label, ps in zip(y_true, y_hat)]
    return np.mean(np.asarray(set_with_corr_label))


def prediction_interval_coverage(
    y_true: Union[np.ndarray, pd.Series],
    prediction_intervals: pd.DataFrame,
    alphas: List[float],
) -> Dict[float, float]:
    """
    Calculate the percentage of prediction intervals that cover the true value.

    This check can be used to validate that the model covers the true values
    according to the alpha set during the prediction process.
    :param y_true: Ground truth labels.
    :param prediction_intervals: DataFrame with predicted bands according to provided confidence levels. This must
        contain columns of format f"{alpha}_low" and f"{1-alpha}_high" for each format.
    :param alphas: List of alphas indicating which confidence levels to check
    """
    if isinstance(y_true, pd.Series):
        y_true = y_true.values

    coverages = {}
    for alpha in alphas:
        coverages[alpha] = np.mean(
            np.where(
                (
                    (prediction_intervals[f"{alpha}_low"] <= y_true)
                    & (prediction_intervals[f"{1-alpha}_high"] >= y_true)
                ),
                1,
                0,
            )
        )

    return coverages



"""Xgboost classification model.

This module contains a wrapper for the Xgboost classification model. It can be used to train and/or tune the model.
It also calculates class weights for imbalanced datasets. The weights may or may not be used deepending on the
hyperparameter tuning.
"""

import logging
from copy import deepcopy
from datetime import datetime
from typing import Any, Dict, List, Literal, Optional, Union

import numpy as np
import optuna
import pandas as pd
import xgboost as xgb
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import KFold, StratifiedKFold
from sklearn.preprocessing import LabelEncoder

from bluecast.config.training_config import (
    TrainingConfig,
    XgboostRegressionFinalParamConfig,
    XgboostTuneParamsRegressionConfig,
)
from bluecast.evaluation.eval_metrics import RegressionEvalWrapper
from bluecast.experimentation.tracking import ExperimentTracker
from bluecast.ml_modelling.base_classes import BaseClassMlRegressionModel
from bluecast.ml_modelling.parameter_tuning_utils import (
    get_params_based_on_device,
    sample_data,
    update_hyperparam_space_after_nth_trial,
    update_params_based_on_tree_method,
    update_params_with_best_params,
)
from bluecast.preprocessing.custom import CustomPreprocessing


class XgboostModelRegression(BaseClassMlRegressionModel):
    """Train and/or tune Xgboost regression model."""

    def __init__(
        self,
        class_problem: Literal["regression"],
        conf_training: Optional[TrainingConfig] = None,
        conf_xgboost: Optional[XgboostTuneParamsRegressionConfig] = None,
        conf_params_xgboost: Optional[XgboostRegressionFinalParamConfig] = None,
        experiment_tracker: Optional[ExperimentTracker] = None,
        custom_in_fold_preprocessor: Optional[CustomPreprocessing] = None,
        cat_columns: Optional[List[Union[str, float, int]]] = None,
        single_fold_eval_metric_func: Optional[RegressionEvalWrapper] = None,
    ):
        self.model: Optional[xgb.XGBRegressor] = None
        self.class_problem = class_problem
        self.conf_training = conf_training
        self.conf_params_xgboost = conf_params_xgboost
        self.conf_xgboost = conf_xgboost

        self.experiment_tracker = experiment_tracker
        self.custom_in_fold_preprocessor = custom_in_fold_preprocessor
        if self.conf_training:
            self.random_generator = np.random.default_rng(
                self.conf_training.global_random_state
            )
        else:
            self.random_generator = np.random.default_rng(0)

        self.cat_columns = cat_columns
        self.single_fold_eval_metric_func = single_fold_eval_metric_func
        self.best_score: float = np.inf

        if not self.single_fold_eval_metric_func:
            self.single_fold_eval_metric_func = RegressionEvalWrapper(
                higher_is_better=False,
                metric_func=mean_squared_error,
                metric_name="Mean squared error",
                **{"squared": False},
            )

    def check_load_confs(self):
        """Load multiple configs or load default configs instead."""
        logging.info("Start loading existing or default config files..")
        if not self.conf_training:
            self.conf_training = TrainingConfig()
            logging.info("Load default TrainingConfig.")
        else:
            logging.info("Found provided TrainingConfig.")

        if not self.conf_xgboost:
            self.conf_xgboost = XgboostTuneParamsRegressionConfig()
            logging.info("Load default XgboostTuneParamsRegressionConfig.")
        else:
            logging.info("Found provided XgboostTuneParamsRegressionConfig.")

        if not self.conf_params_xgboost:
            self.conf_params_xgboost = XgboostRegressionFinalParamConfig()
            logging.info("Load default XgboostRegressionFinalParamConfig.")
        else:
            logging.info("Found provided XgboostRegressionFinalParamConfig.")

    def fit(
        self,
        x_train: pd.DataFrame,
        x_test: pd.DataFrame,
        y_train: pd.Series,
        y_test: pd.Series,
    ) -> xgb.Booster:
        """Train Xgboost model. Includes hyperparameter tuning on default."""
        logging.info("Start fitting Xgboost model.")
        self.check_load_confs()

        if (
            not self.conf_params_xgboost
            or not self.conf_training
            or not self.experiment_tracker
        ):
            raise ValueError(
                "conf_params_xgboost, conf_training or experiment_tracker is None"
            )

        if not self.conf_training.show_detailed_tuning_logs:
            optuna.logging.set_verbosity(optuna.logging.WARNING)

        if self.conf_training.autotune_model:
            self.autotune(x_train, x_test, y_train, y_test)
            print("Finished hyperparameter tuning")

        if self.conf_training.enable_grid_search_fine_tuning:
            self.fine_tune(x_train, x_test, y_train, y_test)
            print("Finished Grid search fine tuning")

        logging.info("Start final model training")
        if self.custom_in_fold_preprocessor:
            x_train, y_train = self.custom_in_fold_preprocessor.fit_transform(
                x_train, y_train
            )
            x_test, y_test = self.custom_in_fold_preprocessor.transform(
                x_test, y_test, predicton_mode=False
            )

        if self.conf_training.use_full_data_for_final_model:
            logging.info(
                f"""{datetime.utcnow()}: Union train and test data for final model training based on TrainingConfig
             param 'use_full_data_for_final_model'"""
            )
            x_train = pd.concat([x_train, x_test]).reset_index(drop=True)
            y_train = pd.concat([y_train, y_test]).reset_index(drop=True)

            if self.cat_columns and self.conf_training.cat_encoding_via_ml_algorithm:
                x_train[self.cat_columns] = x_train[self.cat_columns].astype("category")

        d_train, d_test = self.create_d_matrices(x_train, y_train, x_test, y_test)
        eval_set = [(d_test, "test")]

        steps = self.conf_params_xgboost.params.pop("steps", 300)

        if self.conf_training.early_stopping_rounds and self.conf_xgboost:
            early_stop = xgb.callback.EarlyStopping(
                rounds=self.conf_training.early_stopping_rounds,
                metric_name=self.conf_xgboost.xgboost_eval_metric,
                data_name="test",
                save_best=self.conf_params_xgboost.params["booster"] != "gblinear",
            )
            callbacks = [early_stop]
        else:
            callbacks = None

        if self.conf_training.hypertuning_cv_folds == 1 and self.conf_xgboost:
            self.model = xgb.train(
                self.conf_params_xgboost.params,
                d_train,
                num_boost_round=steps,
                early_stopping_rounds=self.conf_training.early_stopping_rounds,
                evals=eval_set,
                verbose_eval=self.conf_xgboost.verbosity_during_final_model_training,
                callbacks=callbacks,
            )
        elif self.conf_xgboost:
            self.model = xgb.train(
                self.conf_params_xgboost.params,
                d_train,
                num_boost_round=steps,
                early_stopping_rounds=self.conf_training.early_stopping_rounds,
                evals=eval_set,
                verbose_eval=self.conf_xgboost.verbosity_during_final_model_training,
                callbacks=callbacks,
            )
        logging.info("Finished training")
        return self.model

    def autotune(
        self,
        x_train: pd.DataFrame,
        x_test: pd.DataFrame,
        y_train: pd.Series,
        y_test: pd.Series,
    ) -> None:
        """Tune hyperparameters.

        An alternative config can be provided to overwrite the hyperparameter search space.
        """
        logging.info("Start hyperparameter tuning of Xgboost model.")
        if (
            not self.conf_params_xgboost
            or not self.conf_training
            or not self.experiment_tracker
        ):
            raise ValueError(
                "conf_params_xgboost, conf_training or experiment_tracker is None"
            )

        self.check_load_confs()

        if (
            not self.conf_params_xgboost
            or not self.conf_training
            or not self.conf_xgboost
            or not self.experiment_tracker
        ):
            raise ValueError(
                "At least one of the configs or experiment_tracker is None, which is not allowed"
            )

        train_on = get_params_based_on_device(
            self.conf_training, self.conf_params_xgboost, self.conf_xgboost
        )

        x_train, x_test, y_train, y_test = sample_data(
            x_train, x_test, y_train, y_test, self.conf_training
        )

        def objective(trial):
            self.conf_xgboost = update_hyperparam_space_after_nth_trial(
                trial,
                self.conf_xgboost,
                self.conf_training.update_hyperparameter_search_space_after_nth_trial,
            )

            param = {
                "validate_parameters": False,
                "objective": self.conf_xgboost.xgboost_objective,
                "booster": self.conf_xgboost.booster,
                "eval_metric": self.conf_xgboost.xgboost_eval_metric,
                "eta": trial.suggest_float(
                    "eta",
                    self.conf_xgboost.eta_min,
                    self.conf_xgboost.eta_max,
                    log=True,
                ),
                "max_depth": trial.suggest_int(
                    "max_depth",
                    self.conf_xgboost.max_depth_min,
                    self.conf_xgboost.max_depth_max,
                    log=False,
                ),
                "alpha": trial.suggest_float(
                    "alpha",
                    self.conf_xgboost.alpha_min,
                    self.conf_xgboost.alpha_max,
                    log=True,
                ),
                "lambda": trial.suggest_float(
                    "lambda",
                    self.conf_xgboost.lambda_min,
                    self.conf_xgboost.lambda_max,
                    log=True,
                ),
                "gamma": trial.suggest_float(
                    "gamma",
                    self.conf_xgboost.gamma_min,
                    self.conf_xgboost.gamma_max,
                    log=True,
                ),
                "min_child_weight": trial.suggest_float(
                    "min_child_weight",
                    self.conf_xgboost.min_child_weight_min,
                    self.conf_xgboost.min_child_weight_max,
                    log=True,
                ),
                "subsample": trial.suggest_float(
                    "subsample",
                    self.conf_xgboost.sub_sample_min,
                    self.conf_xgboost.sub_sample_max,
                ),
                "colsample_bytree": trial.suggest_float(
                    "colsample_bytree",
                    self.conf_xgboost.col_sample_by_tree_min,
                    self.conf_xgboost.col_sample_by_tree_max,
                ),
                "colsample_bylevel": trial.suggest_float(
                    "colsample_bylevel",
                    self.conf_xgboost.col_sample_by_level_min,
                    self.conf_xgboost.col_sample_by_level_max,
                ),
                "steps": trial.suggest_int(
                    "steps",
                    self.conf_xgboost.steps_min,
                    self.conf_xgboost.steps_max,
                    log=True,
                ),
            }
            params = {**param, **train_on}

            params = update_params_based_on_tree_method(
                params, trial, self.conf_xgboost
            )

            d_train = xgb.DMatrix(
                x_train,
                label=y_train,
                enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
            )

            d_test = xgb.DMatrix(
                x_test,
                label=y_test,
                enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
            )

            pruning_callback = optuna.integration.XGBoostPruningCallback(
                trial, f"test-{self.conf_xgboost.xgboost_eval_metric}"
            )

            steps = params.pop("steps", 300)

            if self.conf_training.hypertuning_cv_folds == 1:
                if self.conf_training.hypertuning_cv_folds == 1:
                    try:
                        return self.train_single_fold_model(
                            d_train, d_test, y_test, params, steps, pruning_callback
                        )
                    except Exception as e:
                        logging.error(f"Error during training: {e}. Pruning trial")
                        trial.should_prune()
            elif (
                self.conf_training.hypertuning_cv_folds > 1
                and self.conf_training.precise_cv_tuning
            ):

                return self._fine_tune_precise(params, x_train, y_train, x_test, y_test)
            else:
                # make regression cv startegy stratified
                le = LabelEncoder()
                y_binned = le.fit_transform(pd.qcut(y_train, 10, duplicates="drop"))
                skf = StratifiedKFold(
                    n_splits=5,
                    random_state=self.conf_training.global_random_state,
                    shuffle=self.conf_training.shuffle_during_training,
                )
                folds = []
                for train_index, test_index in skf.split(x_train, y_binned):
                    folds.append((train_index.tolist(), test_index.tolist()))

                result = xgb.cv(
                    params=params,
                    dtrain=d_train,
                    num_boost_round=steps,
                    # early_stopping_rounds=self.conf_training.early_stopping_rounds, # not recommended as per docs: https://xgboost.readthedocs.io/en/stable/python/sklearn_estimator.html
                    nfold=self.conf_training.hypertuning_cv_folds,
                    as_pandas=True,
                    seed=self.conf_training.global_random_state,
                    callbacks=[pruning_callback],
                    shuffle=self.conf_training.shuffle_during_training,
                    folds=folds,
                )

                adjusted_score = result[
                    f"test-{self.conf_xgboost.xgboost_eval_metric}-mean"
                ].values[-1]

                # track results
                if len(self.experiment_tracker.experiment_id) == 0:
                    new_id = 0
                else:
                    new_id = self.experiment_tracker.experiment_id[-1] + 1
                self.experiment_tracker.add_results(
                    experiment_id=new_id,
                    score_category="cv_score",
                    training_config=self.conf_training,
                    model_parameters=params,
                    eval_scores=adjusted_score,
                    metric_used="adjusted rmse",
                    metric_higher_is_better=False,
                )

                return adjusted_score

        for rst in range(self.conf_training.autotune_n_random_seeds):
            logging.info(
                f"Hyperparameter tuning using random seed {self.conf_training.global_random_state + rst}"
            )

            sampler = optuna.samplers.TPESampler(
                multivariate=True,
                seed=self.conf_training.global_random_state,
                n_startup_trials=self.conf_training.optuna_sampler_n_startup_trials,
                warn_independent_sampling=False,
            )
            study = optuna.create_study(
                direction=self.conf_xgboost.xgboost_eval_metric_tune_direction,
                sampler=sampler,
                study_name="xgboost regression tuning",
                pruner=optuna.pruners.MedianPruner(
                    n_startup_trials=20,
                    n_warmup_steps=20,
                ),
            )

            study.optimize(
                objective,
                n_trials=self.conf_training.hyperparameter_tuning_rounds,
                timeout=self.conf_training.hyperparameter_tuning_max_runtime_secs,
                gc_after_trial=True,
                show_progress_bar=True,
            )

            if self.conf_training.plot_hyperparameter_tuning_overview:
                try:
                    fig = optuna.visualization.plot_optimization_history(study)
                    fig.show()
                    fig = optuna.visualization.plot_param_importances(
                        study  # , evaluator=FanovaImportanceEvaluator()
                    )
                    fig.show()
                except (ZeroDivisionError, RuntimeError, ValueError):
                    pass

            if study.best_value < self.best_score:
                self.best_score = study.best_value
                logging.info(
                    f"New best score: {study.best_value} from random seed  {self.conf_training.global_random_state + rst}"
                )
                xgboost_best_param = study.best_trial.params

                self.conf_params_xgboost.params = {
                    "validate_parameters": False,
                    "objective": self.conf_xgboost.xgboost_objective,  # OR  'binary:logistic' #the loss function being used
                    "booster": self.conf_xgboost.booster,
                    "eval_metric": self.conf_xgboost.xgboost_eval_metric,
                    "max_depth": xgboost_best_param[
                        "max_depth"
                    ],  # maximum depth of the decision trees being trained
                    "alpha": xgboost_best_param["alpha"],
                    "lambda": xgboost_best_param["lambda"],
                    "gamma": xgboost_best_param["gamma"],
                    "min_child_weight": xgboost_best_param["min_child_weight"],
                    "subsample": xgboost_best_param["subsample"],
                    "colsample_bytree": xgboost_best_param["colsample_bytree"],
                    "colsample_bylevel": xgboost_best_param["colsample_bylevel"],
                    "eta": xgboost_best_param["eta"],
                    "steps": xgboost_best_param["steps"],
                }
                self.conf_params_xgboost.params = {
                    **self.conf_params_xgboost.params,
                    **train_on,
                }
                self.conf_params_xgboost.params = update_params_with_best_params(
                    self.conf_params_xgboost.params, xgboost_best_param
                )
                logging.info(f"Best params: {self.conf_params_xgboost.params}")
                print(f"Best params: {self.conf_params_xgboost.params}")

    def create_d_matrices(self, x_train, y_train, x_test, y_test):
        d_train = xgb.DMatrix(
            x_train,
            label=y_train,
            enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
        )
        d_test = xgb.DMatrix(
            x_test,
            label=y_test,
            enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
        )
        return d_train, d_test

    def train_single_fold_model(
        self, d_train, d_test, y_test, param, steps, pruning_callback
    ):
        eval_set = [(d_test, "test")]
        if self.conf_training.early_stopping_rounds and self.conf_xgboost:
            early_stop = xgb.callback.EarlyStopping(
                rounds=self.conf_training.early_stopping_rounds,
                metric_name=self.conf_xgboost.xgboost_eval_metric,
                data_name="test",
                save_best=param["booster"] != "gblinear",
            )
            callbacks = [early_stop]
        else:
            callbacks = None

        model = xgb.train(
            param,
            d_train,
            num_boost_round=steps,
            evals=eval_set,
            callbacks=callbacks,
            verbose_eval=self.conf_xgboost.verbosity_during_hyperparameter_tuning,
        )
        preds = model.predict(d_test)
        mse = self.single_fold_eval_metric_func.regression_eval_func_wrapper(
            y_test, preds
        )

        # track results
        if len(self.experiment_tracker.experiment_id) == 0:
            new_id = 0
        else:
            new_id = self.experiment_tracker.experiment_id[-1] + 1
        self.experiment_tracker.add_results(
            experiment_id=new_id,
            score_category="simple_train_test_score",
            training_config=self.conf_training,
            model_parameters=param,
            eval_scores=mse,
            metric_used="root_mean_squared_error",
            metric_higher_is_better=False,
        )
        return mse

    def _fine_tune_precise(
        self,
        tuned_params: Dict[str, Any],
        x_train: pd.DataFrame,
        y_train: pd.Series,
        x_test: pd.DataFrame,
        y_test: pd.Series,
    ):
        steps = tuned_params.pop("steps", 300)

        if not self.conf_training:
            self.conf_training = TrainingConfig()
            logging.info(
                "Could not find Training config. Falling back to default values"
            )

        stratifier = KFold(
            n_splits=self.conf_training.hypertuning_cv_folds,
            shuffle=True,
            random_state=self.conf_training.global_random_state,
        )

        fold_losses = []
        for _fn, (trn_idx, val_idx) in enumerate(
            stratifier.split(x_train, y_train.astype(int))
        ):
            X_train_fold, X_val_fold = (
                x_train.iloc[trn_idx],
                x_train.iloc[val_idx],
            )
            y_train_fold, y_val_fold = (
                y_train.iloc[trn_idx],
                y_train.iloc[val_idx],
            )
            if self.custom_in_fold_preprocessor:
                (
                    X_train_fold,
                    y_train_fold,
                ) = self.custom_in_fold_preprocessor.fit_transform(
                    X_train_fold, y_train_fold
                )
                (
                    X_test_fold,
                    y_test_fold,
                ) = self.custom_in_fold_preprocessor.transform(x_test, y_test)
                (
                    X_val_fold,
                    y_val_fold,
                ) = self.custom_in_fold_preprocessor.transform(
                    X_val_fold, y_val_fold, predicton_mode=False
                )
            else:
                X_test_fold, y_test_fold = x_test, y_test

            d_test = xgb.DMatrix(
                X_val_fold,
                label=y_val_fold,
                enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
            )

            if not self.conf_params_xgboost:
                self.conf_params_xgboost = XgboostRegressionFinalParamConfig()
                logging.info(
                    "Could not find XgboostRegressionFinalParamConfig. Falling back to default settings."
                )

            d_train = xgb.DMatrix(
                X_train_fold,
                label=y_train_fold,
                enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
            )
            eval_set = [(d_test, "test")]

            if not self.conf_xgboost:
                self.conf_xgboost = XgboostTuneParamsRegressionConfig()
                logging.info(
                    "Could not find XgboostTuneParamsRegressionConfig. Falling back to defaults."
                )

            model = xgb.train(
                tuned_params,
                d_train,
                num_boost_round=steps,
                evals=eval_set,
                verbose_eval=self.conf_xgboost.verbosity_during_hyperparameter_tuning,
            )
            d_eval = xgb.DMatrix(
                X_test_fold,
                label=y_test_fold,
                enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
            )
            preds = model.predict(d_eval)

            if self.single_fold_eval_metric_func:
                loss = self.single_fold_eval_metric_func.regression_eval_func_wrapper(
                    y_test_fold, preds
                )
            else:
                raise ValueError("No single_fold_eval_metric_func could be found")

            fold_losses.append(loss)

        mse_mean = np.mean(np.asarray(fold_losses))

        if self.experiment_tracker and self.conf_training:
            # track results
            if len(self.experiment_tracker.experiment_id) == 0:
                new_id = 0
            else:
                new_id = self.experiment_tracker.experiment_id[-1] + 1
            self.experiment_tracker.add_results(
                experiment_id=new_id,
                score_category="oof_score",
                training_config=self.conf_training,
                model_parameters=tuned_params,
                eval_scores=mse_mean,
                metric_used="root_mean_squared_error",
                metric_higher_is_better=False,
            )
        return mse_mean

    def fine_tune(
        self,
        x_train: pd.DataFrame,
        x_test: pd.DataFrame,
        y_train: pd.Series,
        y_test: pd.Series,
    ) -> None:
        logging.info("Start grid search fine tuning of Xgboost model.")
        if (
            not self.conf_params_xgboost
            or not self.conf_training
            or not self.conf_xgboost
            or not self.experiment_tracker
        ):
            raise ValueError(
                "At least one of the configs or experiment_tracker is None, which is not allowed"
            )

        def objective(trial):
            d_train, d_test = self.create_d_matrices(x_train, y_train, x_test, y_test)

            pruning_callback = optuna.integration.XGBoostPruningCallback(
                trial, f"test-{self.conf_xgboost.xgboost_eval_metric}"
            )
            # copy best params to not overwrite them
            tuned_params = deepcopy(self.conf_params_xgboost.params)
            min_child_weight_space = trial.suggest_float(
                "min_child_weight",
                self.conf_params_xgboost.params["min_child_weight"] * 0.9,
                self.conf_params_xgboost.params["min_child_weight"] * 1.1,
                log=False,
            )
            lambda_space = trial.suggest_float(
                "lambda",
                self.conf_params_xgboost.params["lambda"] * 0.9,
                self.conf_params_xgboost.params["lambda"] * 1.1,
                log=False,
            )
            gamma_space = trial.suggest_float(
                "gamma",
                self.conf_params_xgboost.params["gamma"] * 0.9,
                self.conf_params_xgboost.params["gamma"] * 1.1,
                log=False,
            )
            eta_space = trial.suggest_float(
                "eta",
                self.conf_params_xgboost.params["eta"] * 0.9,
                self.conf_params_xgboost.params["eta"] * 1.1,
                log=False,
            )

            tuned_params["lambda"] = lambda_space
            tuned_params["min_child_weight"] = min_child_weight_space
            tuned_params["gamma"] = gamma_space
            tuned_params["eta"] = eta_space

            steps = tuned_params.pop("steps", 300)

            if self.conf_training.hypertuning_cv_folds == 1:
                try:
                    return self.train_single_fold_model(
                        d_train, d_test, y_test, tuned_params, steps, pruning_callback
                    )
                except Exception as e:
                    logging.error(f"Error during training: {e}. Pruning trial")
                    trial.should_prune()
            elif (
                self.conf_training.hypertuning_cv_folds > 1
                and self.conf_training.precise_cv_tuning
            ):
                return self._fine_tune_precise(
                    tuned_params,
                    x_train,
                    y_train,
                    x_test,
                    y_test,
                )
            else:
                # make regression cv startegy stratified
                le = LabelEncoder()
                y_binned = le.fit_transform(pd.qcut(y_train, 10, duplicates="drop"))
                skf = StratifiedKFold(
                    n_splits=5,
                    random_state=self.conf_training.global_random_state,
                    shuffle=self.conf_training.shuffle_during_training,
                )
                folds = []
                for train_index, test_index in skf.split(x_train, y_binned):
                    folds.append((train_index.tolist(), test_index.tolist()))

                result = xgb.cv(
                    params=tuned_params,
                    dtrain=d_train,
                    num_boost_round=steps,
                    # early_stopping_rounds=self.conf_training.early_stopping_rounds,  # not recommended as per docs: https://xgboost.readthedocs.io/en/stable/python/sklearn_estimator.html
                    nfold=self.conf_training.hypertuning_cv_folds,
                    as_pandas=True,
                    seed=self.conf_training.global_random_state,
                    callbacks=[pruning_callback],
                    shuffle=self.conf_training.shuffle_during_training,
                    folds=folds,
                )

                adjusted_score = result[
                    f"test-{self.conf_xgboost.xgboost_eval_metric}-mean"
                ].values[-1]

                # track results
                if len(self.experiment_tracker.experiment_id) == 0:
                    new_id = 0
                else:
                    new_id = self.experiment_tracker.experiment_id[-1] + 1
                self.experiment_tracker.add_results(
                    experiment_id=new_id,
                    score_category="cv_score",
                    training_config=self.conf_training,
                    model_parameters=tuned_params,
                    eval_scores=adjusted_score,
                    metric_used="root_mean_squared_error",
                    metric_higher_is_better=False,
                )

                return adjusted_score

        self.check_load_confs()
        if (
            isinstance(self.conf_params_xgboost.params["min_child_weight"], float)
            and isinstance(self.conf_params_xgboost.params["lambda"], float)
            and isinstance(self.conf_params_xgboost.params["gamma"], float)
            and isinstance(self.conf_params_xgboost.params["eta"], float)
        ):
            search_space = {
                "min_child_weight": np.linspace(
                    self.conf_params_xgboost.params["min_child_weight"] * 0.9,
                    self.conf_params_xgboost.params["min_child_weight"] * 1.1,
                    self.conf_training.gridsearch_nb_parameters_per_grid,
                    dtype=float,
                ),
                "lambda": np.linspace(
                    self.conf_params_xgboost.params["lambda"] * 0.9,
                    self.conf_params_xgboost.params["lambda"] * 1.1,
                    self.conf_training.gridsearch_nb_parameters_per_grid,
                    dtype=float,
                ),
                "gamma": np.linspace(
                    self.conf_params_xgboost.params["gamma"] * 0.9,
                    self.conf_params_xgboost.params["gamma"] * 1.1,
                    self.conf_training.gridsearch_nb_parameters_per_grid,
                    dtype=float,
                ),
                "eta": np.linspace(
                    self.conf_params_xgboost.params["eta"] * 0.9,
                    self.conf_params_xgboost.params["eta"] * 1.1,
                    self.conf_training.gridsearch_nb_parameters_per_grid,
                    dtype=float,
                ),
            }
        else:
            ValueError("Some parameters are not floats or strings")

        study = optuna.create_study(
            direction=self.conf_xgboost.xgboost_eval_metric_tune_direction,
            sampler=optuna.samplers.GridSampler(search_space),
            pruner=optuna.pruners.MedianPruner(n_startup_trials=20, n_warmup_steps=50),
        )
        study.optimize(
            objective,
            n_trials=self.conf_training.gridsearch_nb_parameters_per_grid
            ** len(search_space.keys()),
            timeout=self.conf_training.gridsearch_tuning_max_runtime_secs,
            gc_after_trial=True,
            show_progress_bar=True,
        )

        if self.conf_training.plot_hyperparameter_tuning_overview:
            try:
                fig = optuna.visualization.plot_optimization_history(study)
                fig.show()
                fig = optuna.visualization.plot_param_importances(
                    study  # , evaluator=FanovaImportanceEvaluator()
                )
                fig.show()
            except (ZeroDivisionError, RuntimeError, ValueError):
                pass

        best_score_cv = self.best_score

        if study.best_value < self.best_score or not self.conf_training.autotune_model:
            self.best_score = study.best_value
            xgboost_grid_best_param = study.best_trial.params
            self.conf_params_xgboost.params["min_child_weight"] = (
                xgboost_grid_best_param["min_child_weight"]
            )
            self.conf_params_xgboost.params["lambda"] = xgboost_grid_best_param[
                "lambda"
            ]
            self.conf_params_xgboost.params["gamma"] = xgboost_grid_best_param["gamma"]
            self.conf_params_xgboost.params["eta"] = xgboost_grid_best_param["eta"]
            logging.info(
                f"Grid search improved eval metric from {best_score_cv} to {self.best_score}."
            )
            logging.info(f"Best params: {self.conf_params_xgboost.params}")
            print(f"Best params: {self.conf_params_xgboost.params}")
        else:
            logging.info(
                f"Grid search could not improve eval metric of {best_score_cv}. Best score reached was {study.best_value}"
            )

    def predict(self, df: pd.DataFrame) -> np.ndarray:
        """Predict on unseen data."""
        logging.info("Start predicting on new data using Xgboost model.")
        if not self.conf_xgboost or not self.conf_training:
            raise ValueError("conf_params_xgboost or conf_training is None")

        if self.custom_in_fold_preprocessor:
            df, _ = self.custom_in_fold_preprocessor.transform(
                df, None, predicton_mode=True
            )

        d_test = xgb.DMatrix(
            df,
            enable_categorical=self.conf_training.cat_encoding_via_ml_algorithm,
        )

        if not self.model:
            raise Exception("No trained model has been found.")

        if not self.conf_params_xgboost:
            raise Exception("No model configuration file has been found.")

        preds = self.model.predict(d_test)
        logging.info("Finished predicting")
        return preds



from typing import Tuple

import numpy as np
import pandas as pd
import pytest

from bluecast.conformal_prediction.effectiveness_nonconformity_measures import (
    avg_c,
    convert_expected_effectiveness_nonconformity_input_types,
    one_c,
)


def create_synthetic_binary_data() -> Tuple[np.ndarray, pd.Series]:
    y_true_binary = pd.Series([0, 1, 0])
    synthetic_results_binary = np.asarray([0.7, 0.3, 0.0])
    return synthetic_results_binary, y_true_binary


def create_synthetic_prediction_set() -> pd.DataFrame:
    synthetic_results_sets = pd.DataFrame(
        {
            "prediction_set": [
                {0},
                {0, 1, 2},
                {0},
            ]
        }
    )
    return synthetic_results_sets


def test_one_c():
    synthetic_results_sets = create_synthetic_prediction_set()
    assert one_c(synthetic_results_sets) == 2 / 3
    assert one_c(pd.DataFrame(synthetic_results_sets)) == 2 / 3
    assert one_c(pd.Series(synthetic_results_sets.values.tolist())) == 2 / 3


def test_avg_c():
    synthetic_results_sets = create_synthetic_prediction_set()
    assert avg_c(synthetic_results_sets) == 5 / 3
    assert avg_c(pd.DataFrame(synthetic_results_sets)) == 5 / 3
    assert avg_c(pd.Series(synthetic_results_sets.values.tolist())) == 5 / 3


def test_convert_expected_effectiveness_nonconformity_input_types():
    synthetic_results_sets = create_synthetic_prediction_set()
    trans_synthetic_results_sets = (
        convert_expected_effectiveness_nonconformity_input_types(
            synthetic_results_sets["prediction_set"]
        )
    )
    assert isinstance(trans_synthetic_results_sets, np.ndarray)

    trans_synthetic_results_sets = (
        convert_expected_effectiveness_nonconformity_input_types(
            synthetic_results_sets["prediction_set"].values
        )
    )
    assert isinstance(trans_synthetic_results_sets, np.ndarray)

    with pytest.raises(ValueError):
        convert_expected_effectiveness_nonconformity_input_types(["a", "b", "c"])



"""Module for evaluation metrics.

This is called as part of the fit_eval function.
"""

import logging
import warnings
from typing import Any, Dict, Literal, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.metrics import (
    accuracy_score,
    classification_report,
    confusion_matrix,
    f1_score,
    log_loss,
    matthews_corrcoef,
    mean_absolute_error,
    mean_squared_error,
    median_absolute_error,
    r2_score,
    recall_score,
    roc_auc_score,
    roc_curve,
)


def plot_lift_chart(y_probs: np.array, y_true: np.array, num_bins: int = 20) -> None:
    # Check if the length of the predicted probabilities matches the length of actual outcomes
    if len(y_probs) != len(y_true):
        raise ValueError(
            "The length of predicted probabilities and actual outcomes must be the same."
        )

    # Create a DataFrame for easy sorting
    data = pd.DataFrame({"Predicted Probability": y_probs, "Actual Outcome": y_true})
    data["Bucket"] = pd.qcut(data["Predicted Probability"], num_bins, duplicates="drop")

    # Calculate the lift values for each bucket
    lift_values = data.groupby("Bucket").agg({"Actual Outcome": ["count", "sum"]})
    lift_values.columns = ["Total Count", "Positive Count"]
    lift_values["Negative Count"] = (
        lift_values["Total Count"] - lift_values["Positive Count"]
    )
    lift_values["Bucket Lift"] = (
        lift_values["Positive Count"] / lift_values["Total Count"]
    ).cumsum()

    # Calculate the baseline lift (if predictions were random)
    random_lift = data["Actual Outcome"].sum() / len(data)

    # Create the lift chart
    plt.figure(figsize=(8, 6))
    plt.plot(
        range(1, num_bins + 1),
        lift_values["Bucket Lift"],
        marker="o",
        label="Model Lift",
    )
    plt.axhline(y=random_lift, color="red", linestyle="--", label="Random Lift")
    plt.xlabel("Bucket")
    plt.ylabel("Lift")
    plt.title("Lift Chart")
    plt.legend()
    plt.grid()
    plt.show()


def plot_roc_auc(
    y_true: np.array, predicted_probabilities: np.array, title="ROC Curve"
) -> None:
    """
    Plot the ROC curve and calculate the AUC (Area Under the Curve).

    :param y_true: True labels (0 or 1) for the binary classification problem.
    :param predicted_probabilities: Predicted probabilities for the positive class.
    :param title: Title for the ROC curve plot.
    """

    # Compute the ROC curve
    fpr, tpr, thresholds = roc_curve(y_true, predicted_probabilities)

    # Calculate AUC
    auc = roc_auc_score(y_true, predicted_probabilities)

    # Plot ROC curve
    plt.figure(figsize=(8, 6))
    plt.plot(fpr, tpr, label=f"AUC = {auc:.2f}")
    plt.plot([0, 1], [0, 1], "k--")
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.0])
    plt.xlabel("False Positive Rate (FPR)")
    plt.ylabel("True Positive Rate (TPR)")
    plt.title(title)
    plt.legend(loc="lower right")
    plt.show()


def plot_probability_distribution(
    y_probs: np.array, num_bins: int = 20, title: str = "Probability Distribution"
) -> None:
    """
    Plot the distribution of predicted probabilities as a histogram using Matplotlib.

    Parameters:
    :param y_probs: NumPy array of predicted probabilities.
    :param num_bins: Number of bins for the histogram (default is 20).
    :param title: Title for the plot (default is "Probability Distribution").
    """
    # Create a histogram of the probabilities
    plt.hist(y_probs, bins=num_bins, edgecolor="k", alpha=0.7)

    # Set plot labels and title
    plt.xlabel("Probability")
    plt.ylabel("Frequency")
    plt.title(title)

    # Display the plot
    plt.show()


def balanced_log_loss(y_true, y_pred):
    if isinstance(y_true, pd.Series) or isinstance(y_true, pd.DataFrame):
        y_true = y_true.values.reshape(-1)
    if isinstance(y_pred, pd.Series) or isinstance(y_pred, pd.DataFrame):
        y_pred = y_pred.values.reshape(-1)

    assert ((y_true == 0) | (y_true == 1)).all()
    assert len(y_true) == len(y_pred)
    assert y_pred.ndim == 1

    eps = 1e-15
    y_pred = y_pred.clip(eps, 1 - eps)
    l0 = -np.log(1 - y_pred[y_true == 0])
    l1 = -np.log(y_pred[y_true != 0])
    return (l0.mean() + l1.mean()) / 2


def eval_classifier(
    y_true: np.ndarray, y_probs: np.ndarray, y_classes: np.ndarray
) -> Dict[str, Any]:
    try:
        matthews = matthews_corrcoef(y_true, y_classes)
    except Exception:
        matthews = 0

    logging.info(f"The Matthew correlation is {matthews}")
    accuracy = accuracy_score(y_true, y_classes)
    logging.info(f"The accuracy is {accuracy}")
    recall = recall_score(y_true, y_classes, average="weighted")
    logging.info(f"The recall is {recall}")
    f1_score_macro = f1_score(y_true, y_classes, average="macro", zero_division=0)
    logging.info(f"The macro F1 score is {f1_score_macro}")
    f1_score_micro = f1_score(y_true, y_classes, average="micro", zero_division=0)
    logging.info(f"The micro F1 score is {f1_score_micro}")
    f1_score_weighted = f1_score(y_true, y_classes, average="weighted", zero_division=0)
    logging.info(f"The weighted F1 score is {f1_score_weighted}")

    if len(y_probs.shape) == 1:
        bll = balanced_log_loss(y_true, y_probs)
        logging.info(f"The balanced logloss is {bll}")
    else:
        bll = 99
        logging.info("Skip balanced logloss as number of classes is less than 2.")

    if len(y_probs.shape) == 1:
        roc_auc = roc_auc_score(y_true, y_probs)
    else:
        roc_auc = roc_auc_score(y_true, y_probs, multi_class="ovr")
    logging.info(f"The ROC auc score is {roc_auc}")
    logloss = log_loss(y_true, y_probs)
    logging.info(f"The log loss score is {logloss}")

    full_classification_report = classification_report(y_true, y_classes)
    logging.info(full_classification_report)

    if len(y_probs.shape) == 1:
        plot_roc_auc(y_true, y_probs)
        try:
            plot_lift_chart(y_probs, y_true)
        except ValueError:
            warnings.warn(
                """Failed to create lift chart. This indicates an issue with the classifier.
            Check if there is any varance in the predicted probabilities.""",
                stacklevel=2,
            )
        plot_probability_distribution(y_probs)
    else:
        logging.info(f"Skip ROC AUC curve as number of classes is {y_probs.shape[1]}.")

    evaluation_scores = {
        "matthews": matthews,
        "accuracy": accuracy,
        "recall": recall,
        "f1_score_macro": f1_score_macro,
        "f1_score_micro": f1_score_micro,
        "f1_score_weighted": f1_score_weighted,
        "log_loss": log_loss,
        "balanced_logloss": bll,
        "roc_auc": roc_auc,
        "classfication_report": full_classification_report,
        "confusion_matrix": confusion_matrix(y_true, y_classes),
    }
    return evaluation_scores


def eval_regressor(y_true: np.ndarray, y_preds: np.ndarray) -> Dict[str, Any]:
    r2 = r2_score(y_true, y_preds)
    print(f"The R2 score is {r2}")
    mean_absolute_error_score = mean_absolute_error(y_true, y_preds)
    print(f"The MAE score is {mean_absolute_error_score}")
    median_absolute_error_score = median_absolute_error(y_true, y_preds)
    print(f"The Median absolute error score is {median_absolute_error_score}")
    mean_squared_error_score = mean_squared_error(y_true, y_preds, squared=True)
    print(f"The MSE score is {mean_squared_error_score}")
    root_mean_squared_error_score = mean_squared_error(y_true, y_preds, squared=False)
    print(f"The RMSE score is {root_mean_squared_error_score}")

    evaluation_scores = {
        "mae": mean_absolute_error_score,
        "r2_score": r2,
        "MSE": mean_squared_error_score,
        "RMSE": root_mean_squared_error_score,
        "median_absolute_error": median_absolute_error_score,
    }
    return evaluation_scores


class ClassificationEvalWrapper:
    """
    Wrapper to evaluate classification metrics.

    :param metric_func: Function object to calculate the metric.
    :param higher_is_better: Boolean indicating if higher metric values are better.
    :param eval_against: String indicating if the metric should be evaluated against probabilities or classes. Can be
        'probas_all_classes', 'probas_target_class' or 'classes'. For 'probas_all_classes', the metric is calculated
        against the predicted probabilities for all classes. For 'probas_target_class', the metric is calculated against
        the predicted probabilities for the best class. For 'classes', the metric is calculated against the predicted
        classes. This parameter decides how the predictions arrive in the metric function.
    :return: Float value of the metric score.
    """

    def __init__(
        self,
        higher_is_better: bool = True,
        eval_against: Literal[
            "probas_all_classes", "probas_target_class", "classes"
        ] = "classes",
        metric_func=matthews_corrcoef,
        metric_name: str = "Matthews Correlation Coefficient",
        **metric_func_args,
    ):
        self.higher_is_better = higher_is_better
        self.eval_against = eval_against
        self.metric_func = metric_func
        self.metric_name = metric_name
        self.metric_func_args = metric_func_args

        if eval_against not in ["probas_all_classes", "probas_target_class", "classes"]:
            raise ValueError(
                f"Argument eval_against must be one of ['probas_all_classes', 'probas_best_class', 'classes']. However {self.eval_against} has been provided."
            )

    def classification_eval_func_wrapper(
        self,
        y_true: Union[np.ndarray, pd.Series, list],
        y_probs: Union[np.ndarray, pd.Series, list],
    ) -> Union[float, int]:
        """
        Wrapper function to evaluate classification metrics.

        :param y_true: Numpy array of true labels.
        :param y_probs: NumPy array of predicted probabilities.
        :return: Float value of the metric score.
        """
        if not isinstance(y_true, list):
            y_true = y_true.tolist()

        if not isinstance(y_probs, list) and self.eval_against != "probas_target_class":
            y_probs = y_probs.tolist()

        if self.eval_against == "probas_all_classes":
            metric_score = self.metric_func(y_true, y_probs, **self.metric_func_args)
        elif self.eval_against == "probas_target_class":
            y_probs_best_class = np.asarray([line[1] for line in y_probs])
            metric_score = self.metric_func(
                y_true, y_probs_best_class, **self.metric_func_args
            )
        elif self.eval_against == "classes":
            y_classes = np.asarray([np.argmax(line) for line in y_probs])
            metric_score = self.metric_func(y_true, y_classes, **self.metric_func_args)
        else:
            raise ValueError(
                f"Unknown value for eval_against: {self.eval_against}. Possible values are 'probas' or 'classes'"
            )

        if self.higher_is_better:
            return -metric_score
        else:
            return metric_score


class RegressionEvalWrapper:
    """
    Wrapper to evaluate regression metrics.

    :param metric_func: Function object to calculate the metric.
    :param higher_is_better: Boolean indicating if higher metric values are better.
    :return: Float value of the metric score.
    """

    def __init__(
        self,
        higher_is_better: bool = False,
        metric_func=mean_squared_error,
        metric_name: str = "Mean squared error",
        **metric_func_args,
    ):
        self.higher_is_better = higher_is_better
        self.metric_func = metric_func
        self.metric_name = metric_name
        self.metric_func_args = metric_func_args

    def regression_eval_func_wrapper(
        self,
        y_true: Union[np.ndarray, pd.Series, list],
        y_hat: Union[np.ndarray, pd.Series, list],
    ) -> Union[float, int]:
        """
        Wrapper function to evaluate classification metrics.

        :param y_true: Numpy array of true targets.
        :param y_hat: NumPy array of predicted targets.
        :return: Float value of the metric score.
        """
        if not isinstance(y_true, list):
            y_true = y_true.tolist()

        if not isinstance(y_hat, list):
            y_hat = y_hat.tolist()

        metric_score = self.metric_func(y_true, y_hat, **self.metric_func_args)

        if self.higher_is_better:
            return -metric_score
        else:
            return metric_score
